Package: bench
Title: High Precision Timing of R Expressions
Version: 1.1.4
Authors@R: c(
    person("Jim", "Hester", role = "aut"),
    person("Davis", "Vaughan", , "davis@posit.co", role = c("aut", "cre")),
    person("Drew", "Schmidt", role = "ctb",
           comment = "read_proc_file implementation"),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: Tools to accurately benchmark and analyze execution times for
    R expressions.
License: MIT + file LICENSE
URL: https://bench.r-lib.org/, https://github.com/r-lib/bench
BugReports: https://github.com/r-lib/bench/issues
Depends: R (>= 4.0.0)
Imports: glue (>= 1.8.0), methods, pillar (>= 1.10.1), profmem (>=
        0.6.0), rlang (>= 1.1.4), stats, tibble (>= 3.2.1), utils
Suggests: covr, dplyr, forcats, ggbeeswarm, ggplot2 (>= 3.5.1),
        ggridges, parallel, scales, testthat (>= 3.2.3), tidyr (>=
        1.3.1), vctrs (>= 0.6.5), withr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-01-16
Encoding: UTF-8
RoxygenNote: 7.3.2
NeedsCompilation: yes
Packaged: 2025-01-16 22:07:42 UTC; davis
Author: Jim Hester [aut],
  Davis Vaughan [aut, cre],
  Drew Schmidt [ctb] (read_proc_file implementation),
  Posit Software, PBC [cph, fnd]
Maintainer: Davis Vaughan <davis@posit.co>
Repository: CRAN
Date/Publication: 2025-01-16 22:40:07 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-28 14:18:02 UTC; unix
Archs: bench.so.dSYM

Package: bench
Title: High Precision Timing of R Expressions
Version: 1.1.4
Authors@R: c(
    person("Jim", "Hester", role = "aut"),
    person("Davis", "Vaughan", , "davis@posit.co", role = c("aut", "cre")),
    person("Drew", "Schmidt", role = "ctb",
           comment = "read_proc_file implementation"),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: Tools to accurately benchmark and analyze execution times for
    R expressions.
License: MIT + file LICENSE
URL: https://bench.r-lib.org/, https://github.com/r-lib/bench
BugReports: https://github.com/r-lib/bench/issues
Depends: R (>= 4.0.0)
Imports: glue (>= 1.8.0), methods, pillar (>= 1.10.1), profmem (>=
        0.6.0), rlang (>= 1.1.4), stats, tibble (>= 3.2.1), utils
Suggests: covr, dplyr, forcats, ggbeeswarm, ggplot2 (>= 3.5.1),
        ggridges, parallel, scales, testthat (>= 3.2.3), tidyr (>=
        1.3.1), vctrs (>= 0.6.5), withr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-01-16
Encoding: UTF-8
RoxygenNote: 7.3.2
NeedsCompilation: yes
Packaged: 2025-01-16 22:07:42 UTC; davis
Author: Jim Hester [aut],
  Davis Vaughan [aut, cre],
  Drew Schmidt [ctb] (read_proc_file implementation),
  Posit Software, PBC [cph, fnd]
Maintainer: Davis Vaughan <davis@posit.co>
Repository: CRAN
Date/Publication: 2025-01-16 22:40:07 UTC

# Generated by roxygen2: do not edit by hand

S3method("[",bench_bytes)
S3method("[",bench_expr)
S3method("[",bench_mark)
S3method("[",bench_time)
S3method("[[",bench_bytes)
S3method("[[",bench_time)
S3method(Ops,bench_bytes)
S3method(Ops,bench_time)
S3method(Summary,bench_time)
S3method(as.character,bench_bytes)
S3method(as.character,bench_expr)
S3method(as.character,bench_time)
S3method(as_bench_bytes,bench_bytes)
S3method(as_bench_bytes,default)
S3method(as_bench_bytes,numeric)
S3method(as_bench_time,bench_time)
S3method(as_bench_time,default)
S3method(as_bench_time,numeric)
S3method(format,bench_bytes)
S3method(format,bench_expr)
S3method(format,bench_time)
S3method(max,bench_bytes)
S3method(max,bench_time)
S3method(mean,bench_time)
S3method(min,bench_bytes)
S3method(min,bench_time)
S3method(pillar_shaft,bench_bytes)
S3method(pillar_shaft,bench_expr)
S3method(pillar_shaft,bench_time)
S3method(plot,bench_mark)
S3method(print,bench_bytes)
S3method(print,bench_expr)
S3method(print,bench_time)
S3method(rbind,bench_mark)
S3method(sum,bench_bytes)
S3method(sum,bench_time)
S3method(summary,bench_mark)
S3method(type_sum,bench_bytes)
S3method(type_sum,bench_expr)
S3method(type_sum,bench_time)
export(as_bench_bytes)
export(as_bench_mark)
export(as_bench_time)
export(bench_bytes)
export(bench_bytes_trans)
export(bench_load_average)
export(bench_memory)
export(bench_process_memory)
export(bench_time)
export(bench_time_trans)
export(hires_time)
export(mark)
export(press)
export(scale_color_bench_expr)
export(scale_colour_bench_expr)
export(scale_x_bench_bytes)
export(scale_x_bench_expr)
export(scale_x_bench_time)
export(scale_y_bench_bytes)
export(scale_y_bench_expr)
export(scale_y_bench_time)
export(system_time)
export(workout)
export(workout_expressions)
importFrom(methods,setOldClass)
importFrom(pillar,pillar_shaft)
importFrom(pillar,type_sum)
useDynLib(bench, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# bench

<!-- badges: start -->

[![R build
status](https://github.com/r-lib/bench/workflows/R-CMD-check/badge.svg)](https://github.com/r-lib/bench)
[![Coverage
status](https://codecov.io/gh/r-lib/bench/branch/master/graph/badge.svg)](https://codecov.io/github/r-lib/bench?branch=master)
[![CRAN
status](https://www.r-pkg.org/badges/version/bench)](https://cran.r-project.org/package=bench)
<!-- badges: end -->

The goal of bench is to benchmark code, tracking execution time, memory
allocations and garbage collections.

## Installation

You can install the release version from
[CRAN](https://cran.r-project.org/) with:

``` r
install.packages("bench")
```

Or you can install the development version from
[GitHub](https://github.com/) with:

``` r
# install.packages("remotes")
remotes::install_github("r-lib/bench")
```

## Features

`bench::mark()` is used to benchmark one or a series of expressions, we
feel it has a number of advantages over [alternatives](#alternatives).

-   Always uses the highest precision APIs available for each operating
    system (often nanoseconds).
-   Tracks memory allocations for each expression.
-   Tracks the number and type of R garbage collections per expression
    iteration.
-   Verifies equality of expression results by default, to avoid
    accidentally benchmarking inequivalent code.
-   Has `bench::press()`, which allows you to easily perform and combine
    benchmarks across a large grid of values.
-   Uses adaptive stopping by default, running each expression for a set
    amount of time rather than for a specific number of iterations.
-   Expressions are run in batches and summary statistics are calculated
    after filtering out iterations with garbage collections. This allows
    you to isolate the performance and effects of garbage collection on
    running time (for more details see [Neal
    2014](https://radfordneal.wordpress.com/2014/02/02/inaccurate-results-from-microbenchmark/)).

The times and memory usage are returned as custom objects which have
human readable formatting for display (e.g. `104ns`) and comparisons
(e.g. `x$mem_alloc > "10MB"`).

There is also full support for plotting with
[ggplot2](http://ggplot2.tidyverse.org/) including custom scales and
formatting.

## Usage

### `bench::mark()`

Benchmarks can be run with `bench::mark()`, which takes one or more
expressions to benchmark against each other.

``` r
library(bench)
set.seed(42)
dat <- data.frame(x = runif(10000, 1, 1000), y=runif(10000, 1, 1000))
```

`bench::mark()` will throw an error if the results are not equivalent,
so you don’t accidentally benchmark inequivalent code.

``` r
bench::mark(
  dat[dat$x > 500, ],
  dat[which(dat$x > 499), ],
  subset(dat, x > 500))
#> Error: Each result must equal the first result:
#> `dat[dat$x > 500, ]` does not equal `dat[which(dat$x > 499), ]`
```

Results are easy to interpret, with human readable units.

``` r
bnch <- bench::mark(
  dat[dat$x > 500, ],
  dat[which(dat$x > 500), ],
  subset(dat, x > 500))
bnch
#> # A tibble: 3 × 6
#>   expression                     min   median `itr/sec` mem_alloc `gc/sec`
#>   <bch:expr>                <bch:tm> <bch:tm>     <dbl> <bch:byt>    <dbl>
#> 1 dat[dat$x > 500, ]           258µs    383µs     2543.     377KB     17.1
#> 2 dat[which(dat$x > 500), ]    204µs    260µs     3803.     260KB     17.7
#> 3 subset(dat, x > 500)         332µs    426µs     2318.     510KB     20.7
```

By default the summary uses absolute measures, however relative results
can be obtained by using `relative = TRUE` in your call to
`bench::mark()` or calling `summary(relative = TRUE)` on the results.

``` r
summary(bnch, relative = TRUE)
#> # A tibble: 3 × 6
#>   expression                  min median `itr/sec` mem_alloc `gc/sec`
#>   <bch:expr>                <dbl>  <dbl>     <dbl>     <dbl>    <dbl>
#> 1 dat[dat$x > 500, ]         1.26   1.47      1.10      1.45     1   
#> 2 dat[which(dat$x > 500), ]  1      1         1.64      1        1.03
#> 3 subset(dat, x > 500)       1.63   1.64      1         1.96     1.21
```

### `bench::press()`

`bench::press()` is used to run benchmarks against a grid of parameters.
Provide setup and benchmarking code as a single unnamed argument then
define sets of values as named arguments. The full combination of values
will be expanded and the benchmarks are then *pressed* together in the
result. This allows you to benchmark a set of expressions across a wide
variety of input sizes, perform replications and other useful tasks.

``` r
set.seed(42)

create_df <- function(rows, cols) {
  as.data.frame(setNames(
    replicate(cols, runif(rows, 1, 100), simplify = FALSE),
    rep_len(c("x", letters), cols)))
}

results <- bench::press(
  rows = c(1000, 10000),
  cols = c(2, 10),
  {
    dat <- create_df(rows, cols)
    bench::mark(
      min_iterations = 100,
      bracket = dat[dat$x > 500, ],
      which = dat[which(dat$x > 500), ],
      subset = subset(dat, x > 500)
    )
  }
)
#> Running with:
#>    rows  cols
#> 1  1000     2
#> 2 10000     2
#> 3  1000    10
#> 4 10000    10
results
#> # A tibble: 12 × 8
#>    expression  rows  cols      min   median `itr/sec` mem_alloc `gc/sec`
#>    <bch:expr> <dbl> <dbl> <bch:tm> <bch:tm>     <dbl> <bch:byt>    <dbl>
#>  1 bracket     1000     2   25.3µs     33µs    29008.   15.84KB    11.6 
#>  2 which       1000     2   25.5µs   32.5µs    29813.    7.91KB     8.95
#>  3 subset      1000     2   45.8µs   56.7µs    17164.    27.7KB     8.66
#>  4 bracket    10000     2   45.6µs     60µs    16286.  156.46KB    56.1 
#>  5 which      10000     2   40.4µs   42.9µs    20583.   78.23KB    31.7 
#>  6 subset     10000     2   94.6µs  117.5µs     8158.  273.79KB    49.1 
#>  7 bracket     1000    10   64.4µs     77µs    12601.   47.52KB    12.8 
#>  8 which       1000    10   58.9µs   63.3µs    14338.    7.91KB    12.4 
#>  9 subset      1000    10   85.1µs  104.7µs     8755.   59.38KB    10.7 
#> 10 bracket    10000    10  128.9µs  144.6µs     6752.   469.4KB    70.3 
#> 11 which      10000    10   89.8µs   97.3µs     9699.   78.23KB    14.8 
#> 12 subset     10000    10  189.5µs  232.9µs     4180.  586.73KB    56.8
```

## Plotting

`ggplot2::autoplot()` can be used to generate an informative default
plot. This plot is colored by gc level (0, 1, or 2) and faceted by
parameters (if any). By default it generates a
[beeswarm](https://github.com/eclarke/ggbeeswarm#geom_quasirandom) plot,
however you can also specify other plot types (`jitter`, `ridge`,
`boxplot`, `violin`). See `?autoplot.bench_mark` for full details.

``` r
ggplot2::autoplot(results)
```

<img src="man/figures/README-autoplot-1.png" width="100%" />

You can also produce fully custom plots by un-nesting the results and
working with the data directly.

## `system_time()`

**bench** also includes `system_time()`, a higher precision alternative
to
[system.time()](https://www.rdocumentation.org/packages/base/versions/3.5.0/topics/system.time).

``` r
bench::system_time({ i <- 1; while(i < 1e7) i <- i + 1 })
#> process    real 
#>   2.37s   2.37s
bench::system_time(Sys.sleep(.5))
#> process    real 
#>    91µs   500ms
```

## Alternatives

-   [rbenchmark](https://cran.r-project.org/package=rbenchmark)
-   [microbenchmark](https://cran.r-project.org/package=microbenchmark)
-   [tictoc](https://cran.r-project.org/package=tictoc)
-   [system.time()](https://www.rdocumentation.org/packages/base/versions/3.5.0/topics/system.time)

NAME
    Bench - Benchmark running times of Perl code

VERSION
    version 0.04

SYNOPSIS
     # time the whole program
     % perl -MBench -e'...'
     0.0123s

     # basic usage of bench()
     % perl -MBench -e'bench sub { ... }'
     100 calls (58548/s), 0.0017s (0.0171ms/call)

     # get bench result in a variable
     % perl -MBench -E'my $res = bench sub { ... }'

     # specify bench options
     % perl -MBench -E'bench sub { ... }, 100'
     % perl -MBench -E'bench sub { ... }, {n=>-5}'
     304347 calls (60665/s), 5.017s (0.0165ms/call)

     # use Dumbbench as the backend
     % perl -MDumbbench -MBench -E'bench sub { ... }'
     % perl -MBench -E'bench sub { ... }, {dumbbench=>1, dumbbench_options=>{...}}'
     Ran 26 iterations (6 outliers).
     Rounded run time per iteration: 2.9029e-02 +/- 4.8e-05 (0.2%)

     # bench multiple codes
     % perl -MBench -E'bench {a=>sub{...}, b=>sub{...}}, {n=>-2}'
     % perl -MBench -E'bench [sub{...}, sub{...}]'; # automatically named a, b, ...
     a: 100 calls (12120/s), 0.0083s (0.0825ms/call)
     b: 100 calls (5357/s), 0.0187s (0.187ms/call)

DESCRIPTION
    This module is an alternative to Benchmark. It provides some nice
    defaults and a simpler interface. There is only one function, bench(),
    and it is exported by default. If bench() is never called, the whole
    program will be timed.

    This module can utilize Dumbbench as the backend.

FUNCTIONS
  bench SUB(S)[, OPTS] => RESULT
    Run Perl code and time it. Exported by default. Will print the result if
    called in void context. SUB can be a coderef for specifying a single
    sub, or hashref/arrayref for specifying multiple subs.

    Options are specified in hashref OPTS. Available options:

    *   n => INT

        Run the code "n" times, or if negative, until at least "n" seconds.

        If unspecified, the default behaviour is to run at most 1 second or
        100 times.

    *   dumbbench => BOOL

        If 0, do not use Dumbbench even if it is available. If 1, require
        and use Dumbbench. If left undef, will use Dumbbench if it is
        already loaded.

    *   dumbbench_options => HASHREF

        Options that will be passed to Dumbbench constructor, e.g.
        {target_rel_precision=>0.005, initial_runs=>20}.

SEE ALSO
    Benchmark

    Dumbbench

AUTHOR
    Steven Haryanto <stevenharyanto@gmail.com>

COPYRIGHT AND LICENSE
    This software is copyright (c) 2012 by Steven Haryanto.

    This is free software; you can redistribute it and/or modify it under
    the same terms as the Perl 5 programming language system itself.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# bench

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/bench)](https://cran.r-project.org/package=bench)
[![R-CMD-check](https://github.com/r-lib/bench/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/bench/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/bench/graph/badge.svg)](https://app.codecov.io/gh/r-lib/bench)

<!-- badges: end -->

The goal of bench is to benchmark code, tracking execution time, memory
allocations and garbage collections.

## Installation

You can install the release version from
[CRAN](https://cran.r-project.org/) with:

``` r
install.packages("bench")
```

Or you can install the development version from
[GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("r-lib/bench")
```

## Features

`bench::mark()` is used to benchmark one or a series of expressions, we
feel it has a number of advantages over [alternatives](#alternatives).

- Always uses the highest precision APIs available for each operating
  system (often nanoseconds).
- Tracks memory allocations for each expression.
- Tracks the number and type of R garbage collections per expression
  iteration.
- Verifies equality of expression results by default, to avoid
  accidentally benchmarking inequivalent code.
- Has `bench::press()`, which allows you to easily perform and combine
  benchmarks across a large grid of values.
- Uses adaptive stopping by default, running each expression for a set
  amount of time rather than for a specific number of iterations.
- Expressions are run in batches and summary statistics are calculated
  after filtering out iterations with garbage collections. This allows
  you to isolate the performance and effects of garbage collection on
  running time (for more details see [Neal
  2014](https://radfordneal.wordpress.com/2014/02/02/inaccurate-results-from-microbenchmark/)).

The times and memory usage are returned as custom objects which have
human readable formatting for display (e.g. `104ns`) and comparisons
(e.g. `x$mem_alloc > "10MB"`).

There is also full support for plotting with
[ggplot2](https://ggplot2.tidyverse.org/) including custom scales and
formatting.

## Usage

### `bench::mark()`

Benchmarks can be run with `bench::mark()`, which takes one or more
expressions to benchmark against each other.

``` r
library(bench)
set.seed(42)

dat <- data.frame(
  x = runif(10000, 1, 1000),
  y = runif(10000, 1, 1000)
)
```

`bench::mark()` will throw an error if the results are not equivalent,
so you don’t accidentally benchmark inequivalent code.

``` r
bench::mark(
  dat[dat$x > 500, ],
  dat[which(dat$x > 499), ],
  subset(dat, x > 500)
)
#> Error: Each result must equal the first result:
#> `dat[dat$x > 500, ]` does not equal `dat[which(dat$x > 499), ]`
```

Results are easy to interpret, with human readable units.

``` r
bnch <- bench::mark(
  dat[dat$x > 500, ],
  dat[which(dat$x > 500), ],
  subset(dat, x > 500)
)
bnch
#> # A tibble: 3 × 6
#>   expression                     min   median `itr/sec` mem_alloc `gc/sec`
#>   <bch:expr>                <bch:tm> <bch:tm>     <dbl> <bch:byt>    <dbl>
#> 1 dat[dat$x > 500, ]           151µs    202µs     4994.     377KB     36.8
#> 2 dat[which(dat$x > 500), ]    124µs    168µs     6169.     260KB     33.2
#> 3 subset(dat, x > 500)         184µs    232µs     3911.     510KB     40.3
```

By default the summary uses absolute measures, however relative results
can be obtained by using `relative = TRUE` in your call to
`bench::mark()` or calling `summary(relative = TRUE)` on the results.

``` r
summary(bnch, relative = TRUE)
#> # A tibble: 3 × 6
#>   expression                  min median `itr/sec` mem_alloc `gc/sec`
#>   <bch:expr>                <dbl>  <dbl>     <dbl>     <dbl>    <dbl>
#> 1 dat[dat$x > 500, ]         1.22   1.20      1.28      1.45     1.11
#> 2 dat[which(dat$x > 500), ]  1      1         1.58      1        1   
#> 3 subset(dat, x > 500)       1.48   1.38      1         1.96     1.21
```

### `bench::press()`

`bench::press()` is used to run benchmarks against a grid of parameters.
Provide setup and benchmarking code as a single unnamed argument then
define sets of values as named arguments. The full combination of values
will be expanded and the benchmarks are then *pressed* together in the
result. This allows you to benchmark a set of expressions across a wide
variety of input sizes, perform replications and other useful tasks.

``` r
set.seed(42)

create_df <- function(rows, cols) {
  out <- replicate(cols, runif(rows, 1, 100), simplify = FALSE)
  out <- setNames(out, rep_len(c("x", letters), cols))
  as.data.frame(out)
}

results <- bench::press(
  rows = c(1000, 10000),
  cols = c(2, 10),
  {
    dat <- create_df(rows, cols)
    bench::mark(
      min_iterations = 100,
      bracket = dat[dat$x > 500, ],
      which = dat[which(dat$x > 500), ],
      subset = subset(dat, x > 500)
    )
  }
)
#> Running with:
#>    rows  cols
#> 1  1000     2
#> 2 10000     2
#> 3  1000    10
#> 4 10000    10

results
#> # A tibble: 12 × 8
#>    expression  rows  cols      min   median `itr/sec` mem_alloc `gc/sec`
#>    <bch:expr> <dbl> <dbl> <bch:tm> <bch:tm>     <dbl> <bch:byt>    <dbl>
#>  1 bracket     1000     2     27µs     34µs    27964.   15.84KB     19.6
#>  2 which       1000     2   25.7µs   33.4µs    29553.    7.91KB     17.7
#>  3 subset      1000     2   45.9µs   58.2µs    16793.    27.7KB     17.1
#>  4 bracket    10000     2   64.1µs   70.8µs    13447.  156.46KB     40.5
#>  5 which      10000     2   46.7µs   54.7µs    17586.   78.23KB     23.3
#>  6 subset     10000     2  116.2µs  132.1µs     7228.  273.79KB     40.9
#>  7 bracket     1000    10   77.2µs   85.4µs    11335.   47.52KB     19.9
#>  8 which       1000    10   67.8µs   75.2µs    13073.    7.91KB     23.2
#>  9 subset      1000    10   84.7µs  107.5µs     9281.   59.38KB     18.8
#> 10 bracket    10000    10  130.2µs  169.1µs     5799.   469.4KB     52.2
#> 11 which      10000    10   75.1µs     96µs    10187.   78.23KB     17.4
#> 12 subset     10000    10  222.7µs    253µs     3810.  586.73KB     43.3
```

## Plotting

`ggplot2::autoplot()` can be used to generate an informative default
plot. This plot is colored by gc level (0, 1, or 2) and faceted by
parameters (if any). By default it generates a
[beeswarm](https://github.com/eclarke/ggbeeswarm#geom_quasirandom) plot,
however you can also specify other plot types (`jitter`, `ridge`,
`boxplot`, `violin`). See `?autoplot.bench_mark` for full details.

``` r
ggplot2::autoplot(results)
```

<img src="man/figures/README-autoplot-1.png" width="100%" />

You can also produce fully custom plots by un-nesting the results and
working with the data directly.

``` r
library(tidyverse)

results %>%
  unnest(c(time, gc)) %>%
  filter(gc == "none") %>%
  mutate(expression = as.character(expression)) %>%
  ggplot(aes(x = mem_alloc, y = time, color = expression)) +
  geom_point() +
  scale_color_bench_expr(scales::brewer_pal(type = "qual", palette = 3))
```

<img src="man/figures/README-custom-plot-1.png" width="100%" />

## `system_time()`

**bench** also includes `system_time()`, a higher precision alternative
to
[system.time()](https://www.rdocumentation.org/packages/base/versions/3.5.0/topics/system.time).

``` r
bench::system_time({
  i <- 1
  while (i < 1e7) {
    i <- i + 1
  }
})
#> process    real 
#>   1.73s    1.7s

bench::system_time(Sys.sleep(.5))
#> process    real 
#>    58µs   497ms
```

## Alternatives

- [rbenchmark](https://cran.r-project.org/package=rbenchmark)
- [microbenchmark](https://cran.r-project.org/package=microbenchmark)
- [tictoc](https://cran.r-project.org/package=tictoc)
- [system.time()](https://www.rdocumentation.org/packages/base/versions/3.5.0/topics/system.time)

#' Return the current high-resolution real time.
#'
#' Time is expressed as seconds since some arbitrary time in the past; it
#' is not correlated in any way to the time of day, and thus is not subject to
#' resetting or drifting. The hi-res
#' timer is ideally suited to performance measurement tasks, where cheap,
#' accurate interval timing is required.
#' @export
#' @examples
#' hires_time()
#'
#' # R rounds doubles to 7 digits by default, see greater precision by setting
#' # the digits argument when printing
#' print(hires_time(), digits = 20)
#'
#' # Generally used by recording two times and then subtracting them
#' start <- hires_time()
#' end <- hires_time()
#' elapsed <- end - start
#' elapsed
hires_time <- function() {
  .Call(hires_time_)
}

describe("hires_time", {
  skip_on_cran()

  it("returns hi resolution times", {
    # it is hard to test this, we will just sleep and verify the second time is
    # after the first.
    start <- hires_time()
    Sys.sleep(.1)
    end <- hires_time()
    expect_is(start, "numeric")
    expect_true(end > start)
  })
})

# This is mostly a copy of https://github.com/r-lib/fs/blob/0f5b6191935fe4c862d2e5003655e6c1669f4afd/R/fs_bytes.R
# If I end up needing this in a third package it should probably live in a package somewhere, maybe prettyunits?

byte_units <- c('B' = 1, 'K' = 1024, 'M' = 1024 ^ 2, 'G' = 1024 ^ 3, 'T' = 1024 ^ 4, 'P' = 1024 ^ 5, 'E' = 1024 ^ 6, 'Z' = 1024 ^ 7, 'Y' = 1024 ^ 8)

#' Human readable memory sizes
#'
#' Construct, manipulate and display vectors of byte sizes. These are numeric
#' vectors, so you can compare them numerically, but they can also be compared
#' to human readable values such as '10MB'.
#'
#' These memory sizes are always assumed to be base 1024, rather than 1000.
#'
#' @param x A numeric or character vector. Character representations can use
#'   shorthand sizes (see examples).
#' @examples
#' bench_bytes("1")
#' bench_bytes("1K")
#' bench_bytes("1Kb")
#' bench_bytes("1KiB")
#' bench_bytes("1MB")
#'
#' bench_bytes("1KB") < "1MB"
#'
#' sum(bench_bytes(c("1MB", "5MB", "500KB")))
#' @name bench_bytes
#' @export
as_bench_bytes <- function(x) {
  UseMethod("as_bench_bytes")
}

#' @export
#' @rdname bench_bytes
bench_bytes <- as_bench_bytes

new_bench_bytes <- function(x) {
  structure(x, class = c("bench_bytes", "numeric"))
}

#' @importFrom methods setOldClass
setOldClass(c("bench_bytes", "numeric"), numeric())

#' @export
as_bench_bytes.default <- function(x) {
  x <- as.character(x)
  m <- captures(x, regexpr("^(?<size>[[:digit:].]+)\\s*(?<unit>[KMGTPEZY]?)i?[Bb]?$", x, perl = TRUE))
  m$unit[m$unit == ""] <- "B"
  new_bench_bytes(unname(as.numeric(m$size) * byte_units[m$unit]))
}

#' @export
as_bench_bytes.bench_bytes <- function(x) {
  return(x)
}

#' @export
as_bench_bytes.numeric <- function(x) {
  new_bench_bytes(x)
}

# Adapted from https://github.com/gaborcsardi/prettyunits
#' @export
format.bench_bytes <- function(x, scientific = FALSE, digits = 3, drop0trailing = TRUE, ...) {
  nms <- names(x)

  bytes <- unclass(x)

  unit <- vcapply(x, find_unit, byte_units)
  res <- round(bytes / byte_units[unit], digits = digits)

  ## Zero bytes
  res[bytes == 0] <- 0
  unit[bytes == 0] <- "B"

  ## NA and NaN bytes
  res[is.na(bytes)] <- NA_real_
  res[is.nan(bytes)] <- NaN
  unit[is.na(bytes)] <- ""            # Includes NaN as well

  # Append an extra B to each unit
  large_units <- unit %in% names(byte_units)[-1]
  unit[large_units] <- paste0(unit[large_units], "B")

  res <- format(res, scientific = scientific, digits = digits, drop0trailing = drop0trailing, ...)

  stats::setNames(paste0(res, unit), nms)
}

#' @export
as.character.bench_bytes <- format.bench_bytes

#' @export
print.bench_bytes <- function(x, ...) {
  print(format.bench_bytes(x, ...), quote = FALSE)
}

#' @export
sum.bench_bytes <- function(x, ...) {
  new_bench_bytes(NextMethod())
}

#' @export
min.bench_bytes <- function(x, ...) {
  new_bench_bytes(NextMethod())
}

#' @export
max.bench_bytes <- function(x, ...) {
  new_bench_bytes(NextMethod())
}

#' @export
`[.bench_bytes` <- function(x, i) {
  new_bench_bytes(NextMethod("["))
}

#' @export
`[[.bench_bytes` <- function(x, i) {
  new_bench_bytes(NextMethod("[["))
}

#' @export
# Adapted from Ops.numeric_version
Ops.bench_bytes <- function (e1, e2) {
  if (nargs() == 1L) {
    stop(sprintf("unary '%s' not defined for \"bench_bytes\" objects", .Generic),
      call. = FALSE)
  }

  boolean <- switch(.Generic,
    `+` = TRUE,
    `-` = TRUE,
    `*` = TRUE,
    `/` = TRUE,
    `^` = TRUE,
    `<` = TRUE,
    `>` = TRUE,
    `==` = TRUE,
    `!=` = TRUE,
    `<=` = TRUE,
    `>=` = TRUE,
  FALSE)
  if (!boolean) {
    stop(sprintf("'%s' not defined for \"bench_bytes\" objects", .Generic),
      call. = FALSE)
  }
  e1 <- as_bench_bytes(e1)
  e2 <- as_bench_bytes(e2)
  NextMethod(.Generic)
}

pillar_shaft.bench_bytes <- function(x, ...) {
  pillar::new_pillar_shaft_simple(format.bench_bytes(x), align = "right", ...)
}

type_sum.bench_bytes <- function(x) {
  "bch:byt"
}


#' Benchmark time transformation
#'
#' This both log transforms the times and formats the labels as a `bench_time`
#' object.
#' @inheritParams bench_time_trans
#' @keywords internal
#' @export
bench_bytes_trans <- function(base = 2) {
  if (is.null(base)) {
    return(
      scales::trans_new("bch:byt", as.numeric, as_bench_bytes,
        scales::pretty_breaks(), domain = c(1e-100, Inf)
      )
    )
  }
  trans <- function(x) log(as.numeric(x), base)
  inv <- function(x) as_bench_bytes(base ^ as.numeric(x))

  scales::trans_new(paste0("bch:byt-", format(base)), trans, inv,
    scales::log_breaks(base = base), domain = c(1e-100, Inf))
}

scale_type.bench_bytes <- function(x) "bench_bytes"

#' Position scales for bench_time data
#'
#' Default scales for the `bench_time` class, these are added to plots using
#' `bench_time` objects automatically.
#' @name scale_bench_time
#' @keywords internal
#' @export
scale_x_bench_bytes <- function(base = 10, ...) {
  ggplot2::scale_x_continuous(..., trans = bench_bytes_trans(base = base))
}

#' @rdname scale_bench_time
#' @keywords internal
#' @export
scale_y_bench_bytes <- function(base = 10, ...) {
  ggplot2::scale_y_continuous(..., trans = bench_bytes_trans(base = base))
}

context("test-bytes.R")

describe("as_bench_bytes", {
  it("accepts numeric input unchanged", {
    expect_equal(unclass(as_bench_bytes(123L)), 123L)
    expect_equal(unclass(as_bench_bytes(123)), 123)
  })
  it("accepts bench_byte input unchanged", {
    x <- as_bench_bytes(123)
    expect_equal(as_bench_bytes(x), x)
  })
  it("coerces character input", {
    expect_equal(unclass(as_bench_bytes("1")), 1)
    expect_equal(unclass(as_bench_bytes("1K")), 1024)
    expect_equal(unclass(as_bench_bytes("1M")), 1024 * 1024)
    expect_equal(unclass(as_bench_bytes("10M")), 10 * 1024 * 1024)
    expect_equal(unclass(as_bench_bytes("1G")), 1024 * 1024 * 1024)
  })
})

describe("format.bench_bytes", {
  it("formats bytes under 1024 as whole numbers", {
    expect_equal(format(bench_bytes(0)), "0B")
    expect_equal(format(bench_bytes(1)), "1B")
    expect_equal(format(bench_bytes(1023)), "1023B")
  })
  it("formats bytes 1024 and up as abbreviated numbers", {
    expect_equal(format(bench_bytes(1024)), "1KB")
    expect_equal(format(bench_bytes(1025)), "1KB")
    expect_equal(format(bench_bytes(2^16)), "64KB")
    expect_equal(format(bench_bytes(2^24)), "16MB")
    expect_equal(format(bench_bytes(2^24 + 555555)), "16.5MB")
    expect_equal(format(bench_bytes(2^32)), "4GB")
    expect_equal(format(bench_bytes(2^48)), "256TB")
    expect_equal(format(bench_bytes(2^64)), "16EB")
  })
  it("handles NA and NaN", {
    expect_equal(format(bench_bytes(NA)), "NA")
    expect_equal(format(bench_bytes(NaN)), "NaN")
  })
  it("works with vectors", {
    v <- c(NA, 1, 2^13, 2^20, NaN, 2^15)
    expect_equal(
      format(bench_bytes(v), trim = TRUE),
      c("NA", "1B", "8KB", "1MB", "NaN", "32KB"))

    expect_equal(format(bench_bytes(numeric())), character())
  })
})

describe("sum.bench_bytes", {
  it("sums its input and returns a bench_byte", {
    expect_equal(sum(bench_bytes(0)), new_bench_bytes(0))
    expect_equal(sum(bench_bytes(c(1, 2))), new_bench_bytes(3))
    expect_equal(sum(bench_bytes(c(1, NA))), new_bench_bytes(NA_real_))
  })
})

describe("min.bench_bytes", {
  it("finds minimum input and returns a bench_byte", {
    expect_equal(min(bench_bytes(0)), new_bench_bytes(0))
    expect_equal(min(bench_bytes(c(1, 2))), new_bench_bytes(1))
    expect_equal(min(bench_bytes(c(1, NA))), new_bench_bytes(NA_real_))
  })
})

describe("max.bench_bytes", {
  it("finds maximum input and returns a bench_byte", {
    expect_equal(max(bench_bytes(0)), new_bench_bytes(0))
    expect_equal(max(bench_bytes(c(1, 2))), new_bench_bytes(2))
    expect_equal(max(bench_bytes(c(1, NA))), new_bench_bytes(NA_real_))
  })
})

describe("[.bench_bytes", {
  it("retains the bench_bytes class", {
    x <- bench_bytes(c(100, 200, 300))
    expect_equal(x[], x)
    expect_equal(x[1], new_bench_bytes(100))
    expect_equal(x[1:2], new_bench_bytes(c(100, 200)))
  })
})

describe("Ops.bench_bytes", {
  it("errors for unary operators", {
    x <- bench_bytes(c(100, 200, 300))
    expect_error(!x, "unary '!' not defined for \"bench_bytes\" objects")
    expect_error(+x, "unary '\\+' not defined for \"bench_bytes\" objects")
    expect_error(-x, "unary '-' not defined for \"bench_bytes\" objects")
  })

  it("works with boolean comparison operators", {
    x <- bench_bytes(c(100, 200, 300))

    expect_equal(x == 100, c(TRUE, FALSE, FALSE))
    expect_equal(x != 100, c(FALSE, TRUE, TRUE))
    expect_equal(x > 100, c(FALSE, TRUE, TRUE))
    expect_equal(x >= 100, c(TRUE, TRUE, TRUE))
    expect_equal(x < 200, c(TRUE, FALSE, FALSE))
    expect_equal(x <= 200, c(TRUE, TRUE, FALSE))
  })

  it("works with arithmetic operators", {
    x <- bench_bytes(c(100, 200, 300))

    expect_equal(x + 100, bench_bytes(c(200, 300, 400)))
    expect_equal(x - 100, bench_bytes(c(0, 100, 200)))
    expect_equal(x * 100, bench_bytes(c(10000, 20000, 30000)))
    expect_equal(x / 2, bench_bytes(c(50, 100, 150)))
    expect_equal(x ^ 2, bench_bytes(c(10000, 40000, 90000)))
  })

  it("errors for other binary operators", {
    x <- bench_bytes(c(100, 200, 300))
    expect_error(x %% 2, "'%%' not defined for \"bench_bytes\" objects")
    expect_error(x %/% 2, "'%/%' not defined for \"bench_bytes\" objects")
    expect_error(x & TRUE, "'&' not defined for \"bench_bytes\" objects")
    expect_error(x | TRUE, "'|' not defined for \"bench_bytes\" objects")
  })
})

#nocov start
.onLoad <- function(...) {
  s3_register("tidyr::unnest", "bench_mark")
  s3_register("dplyr::filter", "bench_mark")
  s3_register("dplyr::group_by", "bench_mark")
  s3_register("ggplot2::autoplot", "bench_mark")

  s3_register("ggplot2::scale_type", "bench_expr")
  s3_register("ggplot2::scale_type", "bench_time")
  s3_register("ggplot2::scale_type", "bench_bytes")

  s3_register("knitr::knit_print", "bench_mark")

  s3_register("vctrs::vec_proxy", "bench_expr")
  s3_register("vctrs::vec_restore", "bench_expr")
}
#nocov end

time_units <- function() {
  stats::setNames(
    c(1e-9,
      1e-6,
      if (is_utf8_output()) 1e-6,
      1e-3,
      1,
      60,
      60 * 60,
      60 * 60 * 24,
      60 * 60 * 24 * 7),
    c("ns",
      "us",
      if (is_utf8_output()) "\U00B5s",
      "ms",
      "s",
      "m",
      "h",
      "d",
      "w")
  )
}

#' Human readable times
#'
#' Construct, manipulate and display vectors of elapsed times in seconds. These
#' are numeric vectors, so you can compare them numerically, but they can also
#' be compared to human readable values such as '10ms'.
#'
#' @param x A numeric or character vector. Character representations can use
#'   shorthand sizes (see examples).
#' @examples
#' as_bench_time("1ns")
#' as_bench_time("1")
#' as_bench_time("1us")
#' as_bench_time("1ms")
#' as_bench_time("1s")
#'
#' as_bench_time("100ns") < "1ms"
#'
#' sum(as_bench_time(c("1MB", "5MB", "500KB")))
#' @export
as_bench_time <- function(x) {
  UseMethod("as_bench_time")
}

new_bench_time <- function(x) {
  structure(x, class = c("bench_time", "numeric"))
}
setOldClass(c("bench_time", "numeric"), numeric())

#' @export
as_bench_time.default <- function(x) {
  x <- as.character(x)
  re <- glue::glue("
      ^(?<size>[[:digit:].]+)\\s*(?<unit>{nms}?)$
      ", nms = paste0(names(time_units()), collapse = "|"))

  m <- captures(x, regexpr(re, x, perl = TRUE))
  m$unit[m$unit == ""] <- "s"
  new_bench_time(unname(as.numeric(m$size) * time_units()[m$unit]))
}

#' @export
as_bench_time.bench_time <- function(x) {
  return(x)
}

#' @export
as_bench_time.numeric <- function(x) {
  is_small <- x < 1e-9 & !is.infinite(x) & x != 0
  x[is_small] <- 1e-9

  new_bench_time(x)
}
tolerance <- sqrt(.Machine$double.eps)
find_unit <- function(x, units) {
  if (is.na(x) || is.nan(x) || x <= 0 || is.infinite(x)) {
    return(NA_character_)
  }
  epsilon <- 1 - (x * (1 / units))
  names(
    utils::tail(n = 1,
      which(epsilon < tolerance)))
}

# Adapted from https://github.com/gaborcsardi/prettyunits
# Aims to be consistent with ls -lh, so uses 1024 KiB units, 3 or less digits etc.
#' @export
format.bench_time <- function(x, scientific = FALSE, digits = 3, drop0trailing = TRUE, ...) {
  nms <- names(x)

  # convert negative times to 1ns, this can happen if the minimum calculated
  # overhead is higher than the time.
  x[x < 1e-9 & !is.infinite(x) & x != 0] <- 1e-9

  seconds <- unclass(x)

  unit <- vcapply(x, find_unit, time_units())
  res <- round(seconds / time_units()[unit], digits = digits)

  ## Zero seconds
  res[seconds == 0] <- 0
  unit[seconds == 0] <- ""

  ## NA, NaN, Inf, -Inf, seconds
  res[is.na(seconds)] <- NA_real_
  res[is.nan(seconds)] <- NaN
  res[is.infinite(seconds)] <- Inf
  res[is.infinite(seconds) & seconds < 0] <- -Inf
  unit[is.na(seconds) | is.infinite(seconds)] <- ""

  res <- format(res, scientific = scientific, digits = digits, drop0trailing = drop0trailing, ...)

  stats::setNames(paste0(res, unit), nms)
}

#' @export
as.character.bench_time <- format.bench_time

#' @export
print.bench_time <- function(x, ...) {
  print(format.bench_time(x, ...), quote = FALSE)
}

#' @export
sum.bench_time <- function(x, ...) {
  new_bench_time(NextMethod())
}

#' @export
min.bench_time <- function(x, ...) {
  new_bench_time(NextMethod())
}

#' @export
max.bench_time <- function(x, ...) {
  new_bench_time(NextMethod())
}

#' @export
`[.bench_time` <- function(x, i, ...) {
  new_bench_time(NextMethod("["))
}

#' @export
`[[.bench_time` <- function(x, i, ...) {
  new_bench_time(NextMethod("[["))
}

#' @export
# Adapted from Ops.numeric_version
Ops.bench_time <- function(e1, e2, ...) {
  if (nargs() == 1L) {
    stop(sprintf("unary '%s' not defined for \"bench_time\" objects", .Generic),
      call. = FALSE)
  }

  boolean <- switch(.Generic,
    `+` = TRUE,
    `-` = TRUE,
    `*` = TRUE,
    `/` = TRUE,
    `^` = TRUE,
    `<` = TRUE,
    `>` = TRUE,
    `==` = TRUE,
    `!=` = TRUE,
    `<=` = TRUE,
    `>=` = TRUE,
  FALSE)
  if (!boolean) {
    stop(sprintf("'%s' not defined for \"bench_time\" objects", .Generic),
      call. = FALSE)
  }
  e1 <- as_bench_time(e1)
  e2 <- as_bench_time(e2)
  NextMethod(.Generic)
}

#' @export
Summary.bench_time <- function(..., na.rm = FALSE) {
  new_bench_time(NextMethod(.Generic))
}

#' @export
mean.bench_time <- function(x, ...) {
  new_bench_time(NextMethod(.Generic))
}

pillar_shaft.bench_time <- function(x, ...) {
  pillar::new_pillar_shaft_simple(format.bench_time(x), align = "right", ...)
}

type_sum.bench_time <- function(x) {
  "bch:tm"
}

#' Benchmark time transformation
#'
#' This both log transforms the times and formats the labels as a `bench_time`
#' object.
#' @inheritParams scales::log_trans
#' @keywords internal
#' @export
bench_time_trans <- function(base = 10) {
  if (is.null(base)) {
    return(
      scales::trans_new("bch:tm", as.numeric, as_bench_time,
        scales::pretty_breaks(), domain = c(1e-100, Inf)
      )
    )
  }

  trans <- function(x) log(as.numeric(x), base)
  inv <- function(x) as_bench_time(base ^ as.numeric(x))

  scales::trans_new(paste0("bch:tm-", format(base)), trans, inv,
    scales::log_breaks(base = base), domain = c(1e-100, Inf))
}

scale_type.bench_time <- function(x) "bench_time"

#' Position scales for bench_time data
#'
#' Default scales for the `bench_time` class, these are added to plots using
#' `bench_time` objects automatically.
#' @name scale_bench_time
#' @param base The base of the logarithm, if `NULL` instead use a
#'   non-logarithmic scale.
#' @keywords internal
#' @export
scale_x_bench_time <- function(base = 10, ...) {
  ggplot2::scale_x_continuous(..., trans = bench_time_trans(base = base))
}

#' @rdname scale_bench_time
#' @keywords internal
#' @export
scale_y_bench_time <- function(base = 10, ...) {
  ggplot2::scale_y_continuous(..., trans = bench_time_trans(base = base))
}

context("test-time.R")

describe("as_bench_time", {
  it("accepts numeric input unchanged", {
    expect_equal(unclass(as_bench_time(123L)), 123L)
    expect_equal(unclass(as_bench_time(123)), 123)
  })
  it("accepts bench_byte input unchanged", {
    x <- as_bench_time(123)
    expect_equal(as_bench_time(x), x)
  })
  it("coerces character input", {
    expect_equal(unclass(as_bench_time("1")), 1)
    expect_equal(unclass(as_bench_time("1ns")), 1e-9)
    expect_equal(unclass(as_bench_time("1us")), 1e-6)
    expect_equal(unclass(as_bench_time("1ms")), 1e-3)
    expect_equal(unclass(as_bench_time("1s")), 1)
    expect_equal(unclass(as_bench_time("1m")), 60)
    expect_equal(unclass(as_bench_time("1h")), 60 * 60)
    expect_equal(unclass(as_bench_time("1d")), 60 * 60 * 24)
    expect_equal(unclass(as_bench_time("1w")), 60 * 60 * 24 * 7)
  })
})

describe("format.as_bench_time", {
  it("formats times under 60 as whole numbers", {
    expect_equal(format(as_bench_time(59)), "59s")
    expect_equal(format(as_bench_time(1)), "1s")
  })
  it("formats times 60 and up as abbreviated minutes / hours / days", {
    withr::with_options(list("cli.unicode" = FALSE), {
      expect_equal(format(as_bench_time(.000000005)), "5ns")
      expect_equal(format(as_bench_time(.0000005)), "500ns")
      expect_equal(format(as_bench_time(.000005)), "5us")
      expect_equal(format(as_bench_time(.0005)), "500us")
      expect_equal(format(as_bench_time(.005)), "5ms")
      expect_equal(format(as_bench_time(.5)), "500ms")
      expect_equal(format(as_bench_time(30)), "30s")
      expect_equal(format(as_bench_time(60)), "1m")
      expect_equal(format(as_bench_time(90)), "1.5m")
      expect_equal(format(as_bench_time(90 * 60)), "1.5h")
      expect_equal(format(as_bench_time(60 * 60 * 60)), "2.5d")
      expect_equal(format(as_bench_time(10.5 * 24 * 60 * 60)), "1.5w")
    })
  })
  it("handles NA and NaN and Inf", {
    expect_equal(format(as_bench_time(NA)), "NA")
    expect_equal(format(as_bench_time(NaN)), "NaN")
    expect_equal(format(as_bench_time(Inf)), "Inf")
    expect_equal(format(as_bench_time(-Inf)), "-Inf")
  })
  it("works with vectors", {
    v <- c(NA, .001, 60, 600, NaN, 6000)
    expect_equal(
      format(as_bench_time(v), trim = TRUE),
      c("NA", "1ms", "1m", "10m", "NaN", "1.67h"))

    expect_equal(format(as_bench_time(numeric())), character())
  })
})

describe("sum.as_bench_time", {
  it("sums its input and returns a bench_byte", {
    expect_equal(sum(as_bench_time(0)), new_bench_time(0))
    expect_equal(sum(as_bench_time(c(1, 2))), new_bench_time(3))
    expect_equal(sum(as_bench_time(c(1, NA))), new_bench_time(NA_real_))
  })
})

describe("min.as_bench_time", {
  it("finds minimum input and returns a bench_byte", {
    expect_equal(min(as_bench_time(0)), new_bench_time(0))
    expect_equal(min(as_bench_time(c(1, 2))), new_bench_time(1))
    expect_equal(min(as_bench_time(c(1, NA))), new_bench_time(NA_real_))
  })
})

describe("max.as_bench_time", {
  it("finds maximum input and returns a bench_byte", {
    expect_equal(max(as_bench_time(0)), new_bench_time(0))
    expect_equal(max(as_bench_time(c(1, 2))), new_bench_time(2))
    expect_equal(max(as_bench_time(c(1, NA))), new_bench_time(NA_real_))
  })
})

describe("[.as_bench_time", {
  it("retains the as_bench_time class", {
    x <- as_bench_time(c(100, 200, 300))
    expect_equal(x[], x)
    expect_equal(x[1], new_bench_time(100))
    expect_equal(x[1:2], new_bench_time(c(100, 200)))
  })
})

describe("Ops.as_bench_time", {
  it("errors for unary operators", {
    x <- as_bench_time(c(100, 200, 300))
    expect_error(!x, "unary '!' not defined for \"bench_time\" objects")
    expect_error(+x, "unary '\\+' not defined for \"bench_time\" objects")
    expect_error(-x, "unary '-' not defined for \"bench_time\" objects")
  })

  it("works with boolean comparison operators", {
    x <- as_bench_time(c(100, 200, 300))

    expect_equal(x == 100, c(TRUE, FALSE, FALSE))
    expect_equal(x != 100, c(FALSE, TRUE, TRUE))
    expect_equal(x > 100, c(FALSE, TRUE, TRUE))
    expect_equal(x >= 100, c(TRUE, TRUE, TRUE))
    expect_equal(x < 200, c(TRUE, FALSE, FALSE))
    expect_equal(x <= 200, c(TRUE, TRUE, FALSE))
  })

  it("works with arithmetic operators", {
    x <- as_bench_time(c(100, 200, 300))

    expect_equal(x + 100, as_bench_time(c(200, 300, 400)))
    expect_equal(x - 100, as_bench_time(c(0, 100, 200)))
    expect_equal(x * 100, as_bench_time(c(10000, 20000, 30000)))
    expect_equal(x / 2, as_bench_time(c(50, 100, 150)))
    expect_equal(x ^ 2, as_bench_time(c(10000, 40000, 90000)))
  })

  it("errors for other binary operators", {
    x <- as_bench_time(c(100, 200, 300))
    expect_error(x %/% 2, "'%/%' not defined for \"bench_time\" objects")
    expect_error(x & TRUE, "'&' not defined for \"bench_time\" objects")
    expect_error(x | TRUE, "'|' not defined for \"bench_time\" objects")
  })
})

viapply <- function(x, f, ...) vapply(x, f, integer(1), ...)
vdapply <- function(x, f, ...) vapply(x, f, double(1), ...)
vcapply <- function(x, f, ...) vapply(x, f, character(1), ...)
vlapply <- function(x, f, ...) vapply(x, f, logical(1), ...)

captures <- function(x, m) {
  assert("`x` must be a character", is.character(x))
  assert("`m` must be a match object from `regexpr()`",
    inherits(m, "integer") &&
    all(c("match.length", "capture.start", "capture.length", "capture.names") %in% names(attributes(m))))

  starts <- attr(m, "capture.start")
  strings <- substring(x, starts, starts + attr(m, "capture.length") - 1L)
  res <- data.frame(matrix(strings, ncol = NCOL(starts)), stringsAsFactors = FALSE)
  colnames(res) <- auto_name_vec(attr(m, "capture.names"))
  res[is.na(m) | m == -1, ] <- NA_character_
  res
}

assert <- function(msg, ..., class = "invalid_argument") {
  tests <- unlist(list(...))

  if (!all(tests)) {
    stop(bench_error(msg, class = class))
  }
}

bench_error <- function(msg, class = "invalid_argument") {
  structure(class = c(class, "bench_error", "error", "condition"), list(message = msg))
}

auto_name_vec <- function(names) {
  missing <- names == ""
  if (all(!missing)) {
    return(names)
  }
  names[missing] <- seq_along(names)[missing]
  names
}

utils::globalVariables("_rval_")

with_gcinfo <- function(expr) {
  tf <- tempfile()
  con <- file(tf, "wb")
  sink(con, type = "message")
  {
    old <- gcinfo(TRUE)
    on.exit({
      gcinfo(old)
      sink(NULL, type = "message")
      close(con)
      output <- readLines(tf, warn = FALSE)
      unlink(tf)
      return(output)
    })
    force(expr)
  }
}

deparse_trunc <- function(x, width = getOption("width")) {
  text <- deparse(x, width.cutoff = width)
  if (length(text) == 1 && nchar(text) < width) return(text)

  # Remove any leading spaces
  text <- sub("^[[:space:]]*", "", text)

  # Collapse all together
  glue::glue_collapse(text, " ", width = width)
}

# inlined from https://github.com/r-lib/cli/blob/master/R/utf8.R
is_utf8_output <- function() {
  opt <- getOption("cli.unicode", NULL)
  if (! is.null(opt)) {
    isTRUE(opt)
  } else {
    l10n_info()$`UTF-8` && !is_latex_output()
  }
}

is_latex_output <- function() {
  if (!("knitr" %in% loadedNamespaces())) return(FALSE)
  get("is_latex_output", asNamespace("knitr"))()
}

collapse <- function(x, sep) {
  paste0(x, collapse = sep)
}

lengths <- function(x, use.names = TRUE) {
  viapply(x, length, USE.NAMES = use.names)
}

# check if the new interface is being used
tidyr_new_interface <- function() {
  utils::packageVersion("tidyr") > "0.8.99"
}

#' Run setup code and benchmarks across a grid of parameters
#'
#' @description
#' `press()` is used to run [bench::mark()] across a grid of parameters and
#' then _press_ the results together.
#'
#' The parameters you want to set are given as named arguments and a grid of
#' all possible combinations is automatically created.
#'
#' The code to setup and benchmark is given by one unnamed expression (often
#' delimited by `\{`).
#'
#' If replicates are desired a dummy variable can be used, e.g. `rep = 1:5` for
#' replicates.
#'
#' @param ... If named, parameters to define, if unnamed the expression to run.
#'   Only one unnamed expression is permitted.
#' @param .grid A pre-build grid of values to use, typically a [data.frame] or
#'   [tibble]. This is useful if you only want to use a subset of all possible combinations.
#' @export
#' @examples
#' # Helper function to create a simple data.frame of the specified dimensions
#' create_df <- function(rows, cols) {
#'   as.data.frame(setNames(
#'     replicate(cols, runif(rows, 1, 1000), simplify = FALSE),
#'     rep_len(c("x", letters), cols)))
#' }
#'
#' # Run 4 data sizes across 3 samples with 2 replicates (24 total benchmarks)
#' press(
#'   rows = c(1000, 10000),
#'   cols = c(10, 100),
#'   rep = 1:2,
#'   {
#'     dat <- create_df(rows, cols)
#'     bench::mark(
#'       min_time = .05,
#'       bracket = dat[dat$x > 500, ],
#'       which = dat[which(dat$x > 500), ],
#'       subset = subset(dat, x > 500)
#'     )
#'   }
#' )
press <- function(..., .grid = NULL) {
  args <- rlang::quos(...)

  unnamed <- names(args) == ""

  if (sum(unnamed) < 1) {
    stop("Must supply one unnamed argument", call. = FALSE)
  }

  if (sum(unnamed) > 1) {
    stop("Must supply no more than one unnamed argument", call. = FALSE)
  }

  if (!is.null(.grid)) {
    if (any(!unnamed)) {
      stop("Must supply either `.grid` or named arguments, not both", call. = FALSE)
    }
    parameters <- .grid
  } else {
    parameters <- expand.grid(lapply(args[!unnamed], rlang::eval_tidy), stringsAsFactors = FALSE)
  }

  status <- format(tibble::as_tibble(parameters), n = Inf)
  message(glue::glue("
      Running with:
      {status[[2]]}"))
  eval_one <- function(row) {
    e <- rlang::new_data_mask(new.env(parent = emptyenv()))
    for (col in seq_along(parameters)) {
      var <- names(parameters)[[col]]
      value <- parameters[row, col]
      assign(var, value, envir = e)
    }
    message(status[[row + 3L]])
    rlang::eval_tidy(args[[which(unnamed)]], data = e)
  }

  res <- lapply(seq_len(nrow(parameters)), eval_one)
  rows <- vapply(res, NROW, integer(1))

  if (!all(rows == rows[[1]])) {
    stop("Results must have equal rows", call. = FALSE)
    # TODO: print parameters / results that are unequal?
  }
  res <- do.call(rbind, res)
  parameters <- parameters[rep(seq_len(nrow(parameters)), each = rows[[1]]), , drop = FALSE]
  bench_mark(tibble::as_tibble(cbind(res[1], parameters, res[-1])))
}

describe("press", {
  it("Adds parameters to output", {
    res <- press(
      x = 1,
      mark(1, max_iterations = 10),
      .quiet = TRUE
    )
    expect_equal(colnames(res), c("expression", "x", summary_cols, data_cols))
    expect_equal(nrow(res), 1)

    res2 <- press(
      x = 1:3,
      mark(1, max_iterations = 10),
      .quiet = TRUE
    )
    expect_equal(colnames(res2), c("expression", "x", summary_cols, data_cols))
    expect_equal(nrow(res2), 3)
  })

  it("Outputs status message before evaluating each parameter", {
    expect_snapshot({
      res <- press(x = 1, mark(rep(1, x), max_iterations = 10))
    })
    expect_equal(colnames(res), c("expression", "x", summary_cols, data_cols))
    expect_equal(nrow(res), 1)

    messages <- character()
    withCallingHandlers(
      res <- press(
        x = 1:3,
        mark(rep(1, x), max_iterations = 10)
      ),
      message = function(cnd) {
        messages <<- append(messages, conditionMessage(cnd))
        if (!is.null(findRestart("muffleMessage"))) {
          invokeRestart("muffleMessage")
        }
      }
    )
    expect_snapshot(messages)
    expect_equal(colnames(res), c("expression", "x", summary_cols, data_cols))
    expect_equal(nrow(res), 3)
  })

  it("expands the grid if has named parameters", {
    res <- press(
      x = c(1, 2),
      y = c(1, 3),
      mark(list(x, y), max_iterations = 10),
      .quiet = TRUE
    )

    expect_equal(res$x, c(1, 2, 1, 2))
    expect_equal(res$y, c(1, 1, 3, 3))
    expect_equal(res$result[[1]], list(1, 1))
    expect_equal(res$result[[2]], list(2, 1))
    expect_equal(res$result[[3]], list(1, 3))
    expect_equal(res$result[[4]], list(2, 3))
  })

  it("takes values as-is if given in .grid", {
    res <- press(
      .grid = data.frame(x = c(1, 2), y = c(1, 3)),
      mark(list(x, y), max_iterations = 10),
      .quiet = TRUE
    )

    expect_equal(res$x, c(1, 2))
    expect_equal(res$y, c(1, 3))
    expect_equal(res$result[[1]], list(1, 1))
    expect_equal(res$result[[2]], list(2, 3))
  })

  it("runs `setup` with the parameters evaluated", {
    x <- 1
    res <- press(
      y = 2,
      {
        x <- y
        mark(x)
      },
      .quiet = TRUE
    )

    expect_equal(res$result[[1]], 2)
  })

  it("`.grid` subsets tibbles and data.frames the same way (#142)", {
    x <- data.frame(a = 1, b = 2)
    out <- press(mark(c(a, b)), .grid = x, .quiet = TRUE)
    expect_identical(out$result[[1L]], c(1, 2))

    x <- tibble::tibble(a = 1, b = 2)
    out <- press(mark(c(a, b)), .grid = x, .quiet = TRUE)
    expect_identical(out$result[[1L]], c(1, 2))
  })
})

#' Workout a group of expressions individually
#'
#' Given an block of expressions in `{}` [workout()] individually times each
#' expression in the group. [workout_expressions()] is a lower level function most
#' useful when reading lists of calls from a file.
#'
#' @param expr one or more expressions to workout, use `{}` to pass multiple
#'   expressions.
#' @param exprs A list of calls to measure.
#' @param description A name to label each expression, if not supplied the
#'   deparsed expression will be used.
#' @param env The environment in which the expressions should be evaluated.
#' @export
#' @examples
#' workout({
#'   x <- 1:1000
#'   evens <- x %% 2 == 0
#'   y <- x[evens]
#'   length(y)
#'   length(which(evens))
#'   sum(evens)
#' })
#'
#' # The equivalent to the above, reading the code from a file
#' workout_expressions(as.list(parse(system.file("examples/exprs.R", package = "bench"))))
workout <- function(expr, description = NULL) {
  expr <- rlang::enquo(expr)
  env <- rlang::quo_get_env(expr)
  if (rlang::quo_get_expr(expr)[[1]] == "{") {
    exprs <- as.list(rlang::quo_get_expr(expr)[-1])
  } else {
    exprs <- list(rlang::quo_get_expr(expr))
  }
  workout_expressions(exprs, env, description)
}

#' @rdname workout
#' @export
workout_expressions <- function(exprs, env = parent.frame(), description = NULL) {
  if (is.null(description)) {
    description <- names(exprs)
  }

  out <- list(
    exprs = new_bench_expr(exprs, description),
    process = numeric(length(exprs)),
    real = numeric(length(exprs))
  )

  for (i in seq_along(exprs)) {
    res <- as_bench_time(.Call(system_time_, exprs[[i]], env))
    out[[2]][[i]] <- res[[1]]
    out[[3]][[i]] <- res[[2]]
  }

  out[[2]] <- new_bench_time(out[[2]])
  out[[3]] <- new_bench_time(out[[3]])

  tibble::as_tibble(out)
}

describe("workout", {
  it("times each expression and names them", {

    res <- workout(
      x <- 1:1000
    )
    expect_named(res, c("exprs", "process", "real"))
    expect_true(nrow(res) == 1)

    res2 <- workout({
      x <- 1:1000
      evens <- x %% 2 == 0
      y <- x[evens]
      length(y)
      length(which(evens))
      sum(evens)
    })
    expect_named(res2, c("exprs", "process", "real"))
    expect_true(nrow(res2) == 6)
  })
})

describe("workout_expressions", {
  it("times given expressions", {
    res <- workout_expressions(as.list(parse(file = system.file("examples/exprs.R", package = "bench"))))
    expect_named(res, c("exprs", "process", "real"))
    expect_true(nrow(res) == 6)
  })
})

#' Get system load averages
#'
#' Uses OS system APIs to return the load average for the past 1, 5 and 15 minutes.
#' @export
bench_load_average <- function() {
  stats::setNames(
    .Call(bench_load_average_),
    c("load_1_min", "load_5_min", "load_15_min")
  )
}

new_bench_expr <- function(x, description = names(x)) {
  if (is.null(description)) {
    description <- rep("", length(x))
  }
  names(x) <- description
  structure(x, class = c("bench_expr", "list"), description = description)
}

#' @export
format.bench_expr <- function(x, ...) {
  desc <- attr(x, "description")
  is_missing <- desc == ""
  desc[is_missing] <- vapply(x[is_missing], deparse_trunc, character(1))
  desc
}

#' @export
as.character.bench_expr <- format.bench_expr

#' @export
print.bench_expr <- function(x, ...) {
  x <- unclass(x)
  NextMethod()
}

#' @export
type_sum.bench_expr <- function(x) {
  "bch:expr"
}

#' @export
`[.bench_expr` <- function(x, i, ...) {
  new_x <- NextMethod("[")
  new_bench_expr(new_x)
}

# Lazily registered in `.onLoad()`
vec_proxy.bench_expr <- function(x, ...) {
  desc <- attr(x, "description")
  attributes(x) <- NULL
  out <- list(x = x, desc = desc)
  vctrs::new_data_frame(out, n = length(x))
}

# Lazily registered in `.onLoad()`
vec_restore.bench_expr <- function(x, to, ...) {
  new_bench_expr(x$x, x$desc)
}

#' @export
pillar_shaft.bench_expr <- function(x, ...) {
  # We format bench expressions exactly like character vectors. This ensures
  # they are truncated as needed, which is useful for long unnamed expressions
  # (#94). This is the same logic as `pillar:::pillar_shaft.factor()`.
  pillar_shaft(as.character(x), ...)
}

# Lazily registered in `.onLoad()`
scale_type.bench_expr <- function(x) {
  "bench_expr"
}

setOldClass(c("bench_expr", "list"), list())

#' Position and color scales for bench_expr data
#'
#' Default scales for the `bench_expr` class, these are added to plots using
#' `bench_expr` objects automatically.
#' @name scale_bench_expr
#' @keywords internal
#' @export
scale_x_bench_expr <- function(...) {
  sc <- ggplot2::scale_x_discrete(...)
  sc$transform <- as.character
  sc
}

#' @rdname scale_bench_expr
#' @keywords internal
#' @export
scale_y_bench_expr <- function(...) {
  sc <- ggplot2::scale_y_discrete(...)
  sc$transform <- as.character
  sc
}

#' @rdname scale_bench_expr
#' @keywords internal
#' @export
scale_colour_bench_expr <- function(
  palette = scales::hue_pal(...),
  ...,
  aesthetics = "colour"
) {
  sc <- ggplot2::discrete_scale(aesthetics, "bench_expr", palette, ...)
  sc$transform <- as.character
  sc
}

#' @rdname scale_bench_expr
#' @keywords internal
#' @export
scale_color_bench_expr <- scale_colour_bench_expr

test_that("`description` is sliced along with expressions", {
  x <- as.list(expression(x + y, z + b))
  x <- new_bench_expr(x, c("a", "b"))

  expect_identical(attr(x[2], "description"), "b")
  expect_identical(attr(x[c(2, 2, 1)], "description"), c("b", "b", "a"))
})

test_that("`vec_slice()` slices `description` attribute", {
  skip_if_not_installed("vctrs")

  x <- as.list(expression(x + y, z + b))
  x <- new_bench_expr(x, c("a", "b"))

  expect_identical(attr(vctrs::vec_slice(x, 2), "description"), "b")
  expect_identical(attr(vctrs::vec_slice(x, c(2, 2, 1)), "description"), c("b", "b", "a"))
})

#' Retrieve the current and maximum memory from the R process
#'
#' The memory reported here will likely differ from that reported by `gc()`, as
#' this includes all memory from the R process, including any child processes
#' and memory allocated outside R's garbage collector heap.
#'
#' The OS APIs used are as follows
#'
#' ## Windows
#' - [PROCESS_MEMORY_COUNTERS.WorkingSetSize](https://docs.microsoft.com/en-us/windows/win32/api/psapi/ns-psapi-process_memory_counters)
#' - [PROCESS_MEMORY_COUNTERS.PeakWorkingSetSize](https://docs.microsoft.com/en-us/windows/win32/api/psapi/ns-psapi-process_memory_counters)
#' ## macOS
#' - [task_info(TASK_BASIC_INFO)](https://developer.apple.com/documentation/kernel/1537934-task_info?language=occ)
#' - [rusage.ru_maxrss](https://developer.apple.com/library/archive/documentation/System/Conceptual/ManPages_iPhoneOS/man2/getrusage.2.html)
#' ## linux
#' - [/proc/pid/status VmSize](https://man7.org/linux/man-pages/man5/proc.5.html)
#' - [/proc/pid/status VmPeak](https://man7.org/linux/man-pages/man5/proc.5.html)
#' and on Windows
#' [PROCESS_MEMORY_COUNTERS.PeakWorkingSetSize](https://docs.microsoft.com/en-us/windows/win32/api/psapi/ns-psapi-process_memory_counters)
#' @export
bench_process_memory <- function() {
  stats::setNames(
    as_bench_bytes(.Call(bench_process_memory_)),
    c("current", "max")
  )
}

describe("bench_process_memory", {
  it("has a current and max memory of bench bytes", {
    res <- bench_process_memory()
    expect_named(res, c("current", "max"))
    expect_s3_class(res[["current"]], "bench_bytes")
    expect_s3_class(res[["max"]], "bench_bytes")
  })

  # This test is unreliable due to when gcs happen when run repeatedly, so it
  # is commented out.
  #it("current memory increases when you allocate a medium size vector", {
    #res1 <- bench_process_memory()
    #x <- rep(1, 1e8)
    #res2 <- bench_process_memory()
    #expect_true(res2[["current"]] > res1[["current"]])
  #})
})

#' Measure Process CPU and real time that an expression used.
#'
#' @param expr A expression to be timed.
#' @return A `bench_time` object with two values.
#' - `process` - The process CPU usage of the expression evaluation.
#' - `real` - The wallclock time of the expression evaluation.
#' @details On some systems (such as macOS) the process clock has lower
#'   precision than the realtime clock, as a result there may be cases where the
#'   process time is larger than the real time for fast expressions.
#' @examples
#' # This will use ~.5 seconds of real time, but very little process time.
#' bench_time(Sys.sleep(.5))
#' @seealso [bench_memory()] To measure memory allocations for a given expression.
#' @aliases system_time
#' @export
bench_time <- function(expr) {
  stats::setNames(
    as_bench_time(.Call(system_time_, substitute(expr), parent.frame())),
    c("process", "real"))
}

#' @export
system_time <- bench_time

#' Measure memory that an expression used.
#'
#' @param expr A expression to be measured.
#' @return A tibble with two columns
#' - The total amount of memory allocated
#' - The raw memory allocations as parsed by [profmem::readRprofmem()]
#' @examples
#' if (capabilities("profmem")) {
#'   bench_memory(1 + 1:10000)
#' }
#' @export
bench_memory <- function(expr) {
  can_profile_memory <- capabilities("profmem")
  if (!can_profile_memory) {
    stop("Memory profiling not available in this version of R", call. = FALSE)
  }
  f <- tempfile()
  on.exit(unlink(f))
  utils::Rprofmem(f, threshold = 1)
  force(expr)
  utils::Rprofmem(NULL)

  memory <- parse_allocations(f)

  tibble::tibble(mem_alloc = bench_bytes(sum(memory$bytes, na.rm = TRUE)), memory = list(memory))
}

describe("bench_time", {
  skip_on_cran()

  res <- bench_time(1 + 1:1e7)
  it("returns process and real time", {
    expect_equal(names(res), c("process", "real"))
  })
  it("returns times that are reasonable, system and real time are relatively
    close for process bound expressions", {
    epsilon <- abs(res[[1]] - res[[2]])
    expect_true((epsilon / res[[1]]) < 1)
  })
  it("returns times that are reasonable, system and real time are far apart
    for non-process bound expressions", {
    res <- bench_time(Sys.sleep(.5))
    epsilon <- abs(res[[1]] - res[[2]])
    expect_true((epsilon / res[[1]]) > 100)
  })
})

describe("bench_memory", {
  skip_on_cran()

  res <- bench_memory(1 + 1:1e7)
  it("returns memory allocation and the raw memory used", {
    expect_equal(names(res), c("mem_alloc", "memory"))
  })
  it("returns reasonable memory allocation", {
      expect_true(res[["mem_alloc"]] > "10MB")
    })
})

#' Autoplot method for bench_mark objects
#'
#' @param object A `bench_mark` object.
#' @param type The type of plot. Plotting geoms used for each type are
#' - beeswarm - [ggbeeswarm::geom_quasirandom()]
#' - jitter - [ggplot2::geom_jitter()]
#' - ridge - [ggridges::geom_density_ridges()]
#' - boxplot - [ggplot2::geom_boxplot()]
#' - violin - [ggplot2::geom_violin()]
#' @param ... Additional arguments passed to the plotting geom.
#' @details This function requires some optional dependencies. [ggplot2][ggplot2::ggplot2-package],
#' [tidyr][tidyr::tidyr-package], and depending on the plot type
#' [ggbeeswarm][ggbeeswarm::ggbeeswarm], [ggridges][ggridges::ggridges-package].
#'
#' For `type` of `beeswarm` and `jitter` the points are colored by the highest
#' level garbage collection performed during each iteration.
#'
#' For plots with 2 parameters `ggplot2::facet_grid()` is used to construct a
#' 2d facet. For other numbers of parameters `ggplot2::facet_wrap()` is used
#' instead.
#'
#' @examples
#' dat <- data.frame(x = runif(10000, 1, 1000), y=runif(10000, 1, 1000))
#'
#' res <- bench::mark(
#'   dat[dat$x > 500, ],
#'   dat[which(dat$x > 500), ],
#'   subset(dat, x > 500))
#'
#' if (require(ggplot2) && require(tidyr) && require(ggbeeswarm)) {
#'
#'   # Beeswarm plot
#'   autoplot(res)
#'
#'   # ridge (joyplot)
#'   autoplot(res, "ridge")
#'
#'   # If you want to have the plots ordered by execution time you can do so by
#'   # ordering factor levels in the expressions.
#'   if (require(dplyr) && require(forcats)) {
#'
#'     res %>%
#'       mutate(expression = forcats::fct_reorder(as.character(expression), min, .desc = TRUE)) %>%
#'       as_bench_mark() %>%
#'       autoplot("violin")
#'   }
#' }
# Lazily registered in `.onLoad()`
autoplot.bench_mark <- function(
  object,
  type = c("beeswarm", "jitter", "ridge", "boxplot", "violin"),
  ...
) {
  rlang::check_installed(c("ggplot2", "tidyr (>= 1.0.0)"), "for `autoplot()`.")

  type <- match.arg(type)

  if (type == "beeswarm") {
    rlang::check_installed("ggbeeswarm", "to use `type = \"beeswarm\".")
  }

  # Just convert bench_expr to characters
  if (inherits(object$expression, "bench_expr")) {
    object$expression <- as.character(object$expression)
  }

  res <- tidyr::unnest(object, c(time, gc))
  p <- ggplot2::ggplot(res)

  switch(
    type,
    beeswarm = p <- p +
      ggplot2::aes(.data$time, .data$expression, color = .data$gc) +
      ggbeeswarm::geom_quasirandom(..., orientation = "y"),

    jitter = p <- p +
      ggplot2::aes(.data$time, .data$expression, color = .data$gc) +
      ggplot2::geom_jitter(...),

    ridge = p <- p +
      ggplot2::aes(.data$time, .data$expression) +
      ggridges::geom_density_ridges(...),

    boxplot = p <- p +
      ggplot2::aes(.data$time, .data$expression) +
      ggplot2::geom_boxplot(...),

    violin = p <- p +
      ggplot2::aes(.data$time, .data$expression) +
      ggplot2::geom_violin(...)
  )

  parameters <- setdiff(
    colnames(object),
    c("expression", summary_cols, data_cols, c("level0", "level1", "level2"))
  )

  if (length(parameters) == 0) {
    return(p)
  }

  if (length(parameters) == 2) {
    return(
      p +
        ggplot2::facet_grid(
          paste0(parameters[[1]], "~", parameters[[2]]),
          labeller = ggplot2::label_both
        )
    )
  }

  p + ggplot2::facet_wrap(parameters, labeller = ggplot2::label_both)
}

#' @rdname autoplot.bench_mark
#' @param x A `bench_mark` object.
#' @param y Ignored, required for compatibility with the `plot()` generic.
#' @export
plot.bench_mark <- function(
  x,
  ...,
  type = c("beeswarm", "jitter", "ridge", "boxplot", "violin"),
  y
) {
  type <- match.arg(type)
  ggplot2::autoplot(x, type = type, ...)
}

#' @useDynLib bench, .registration = TRUE
NULL

#' Benchmark a series of functions
#'
#' Benchmark a list of quoted expressions. Each expression will always run at
#' least twice, once to measure the memory allocation and store results and one
#' or more times to measure timing.
#'
#' @param ... Expressions to benchmark, if named the `expression` column will
#'   be the name, otherwise it will be the deparsed expression.
#' @param min_time The minimum number of seconds to run each expression, set to
#'   `Inf` to always run `max_iterations` times instead.
#' @param iterations If not `NULL`, the default, run each expression for
#'   exactly this number of iterations. This overrides both `min_iterations`
#'   and `max_iterations`.
#' @param exprs A list of quoted expressions. If supplied overrides expressions
#'   defined in `...`.
#' @param min_iterations Each expression will be evaluated a minimum of `min_iterations` times.
#' @param max_iterations Each expression will be evaluated a maximum of `max_iterations` times.
#' @param check Check if results are consistent. If `TRUE`, checking is done
#'   with [all.equal()], if `FALSE` checking is disabled and results are not
#'   stored. If `check` is a function that function will be called with each
#'   pair of results to determine consistency.
#' @param memory If `TRUE` (the default when R is compiled with memory
#'   profiling), track memory allocations using. If `FALSE` disable memory
#'   tracking.
#' @param env The environment which to evaluate the expressions
#' @inheritParams summary.bench_mark
#' @inherit summary.bench_mark return
#' @aliases bench_mark
#' @seealso [press()] to run benchmarks across a grid of parameters.
#' @examples
#' dat <- data.frame(x = runif(100, 1, 1000), y=runif(10, 1, 1000))
#' mark(
#'   min_time = .1,
#'
#'   dat[dat$x > 500, ],
#'   dat[which(dat$x > 500), ],
#'   subset(dat, x > 500))
#' @export
mark <- function(..., min_time = .5, iterations = NULL, min_iterations = 1,
                 max_iterations = 10000, check = TRUE, memory = capabilities("profmem"), filter_gc = TRUE,
                 relative = FALSE, time_unit = NULL, exprs = NULL, env = parent.frame()) {

  if (!is.null(iterations)) {
    min_iterations <- iterations
    max_iterations <- iterations
  }

  if (isTRUE(check)) {
    check_fun <- all.equal
  } else if (is.function(check)) {
    check_fun <- check
    check <- TRUE
  } else {
    check <- FALSE
  }

  if (is.null(exprs)) {
    exprs <- dots(...)
  }

  n_exprs <- length(exprs)

  results <- list(expression = new_bench_expr(exprs), time = vector("list", n_exprs), gc = vector("list", n_exprs), memory = vector("list", n_exprs), result = vector("list", n_exprs))

  # Helper for evaluating with memory profiling
  eval_one <- function(e, profile_memory) {
    f <- tempfile()
    on.exit(unlink(f))
    if (profile_memory) {
      utils::Rprofmem(f, threshold = 1)
    }

    res <- eval(e, env)
    if (profile_memory) {
      utils::Rprofmem(NULL)
    }
    list(result = res, memory = parse_allocations(f))
  }

  # We only want to evaluate these first runs if we need to check memory or results.
  if (memory || check) {
    # Run allocation benchmark and check results
    for (i in seq_len(length(exprs))) {
      res <- eval_one(exprs[[i]], memory)
      if (check) {
        if (is.null(res$result)) {
          results$result[i] <- list(res$result)
        } else {
          results$result[[i]] <- res$result
        }
      }
      if (memory) {
        results$memory[[i]] <- res$memory
      }

      if (check && i > 1) {
        comp <- check_fun(results$result[[1]], results$result[[i]])
        if (!isTRUE(comp)) {
          expressions <- as.character(results$expression)

          stop(glue::glue("
              Each result must equal the first result:
              `{first}` does not equal `{current}`
              ",
              first = expressions[[1]],
              current = expressions[[i]]
              ),
            call. = FALSE
          )
        }
      }
    }
  }

  for (i in seq_len(length(exprs))) {
    error <- NULL
    gc_msg <- with_gcinfo({
      tryCatch(error = function(e) { e$call <- NULL; error <<- e},
      res <- .Call(mark_, exprs[[i]], env, min_time, as.integer(min_iterations), as.integer(max_iterations))
      )
    })
    if (!is.null(error)) {
      stop(error)
    }

    results$time[[i]] <- as_bench_time(res)
    results$gc[[i]] <- parse_gc(gc_msg)
  }

  out <- summary(bench_mark(tibble::as_tibble(results, .name_repair = "minimal")),
          filter_gc = filter_gc, relative = relative, time_unit = time_unit)

  out
}

bench_mark <- function(x) {
  class(x) <- unique(c("bench_mark", class(x)))
  x
}

#' Coerce to a bench mark object Bench mark objects
#'
#' This is typically needed only if you are performing additional manipulations
#' after calling [bench::mark()].
#' @param x Object to be coerced
#' @export
as_bench_mark <- function(x) {
  bench_mark(tibble::as_tibble(x))
}

summary_cols <- c("min", "median", "itr/sec", "mem_alloc", "gc/sec")
data_cols <- c("n_itr", "n_gc", "total_time", "result", "memory", "time", "gc")
time_cols <- c("min", "median", "total_time")

#' Summarize [bench::mark] results.
#'
#' @param object [bench_mark] object to summarize.
#' @param filter_gc If `TRUE` remove iterations that contained at least one
#'   garbage collection before summarizing. If `TRUE` but an expression had
#'   a garbage collection in every iteration, filtering is disabled, with a warning.
#' @param relative If `TRUE` all summaries are computed relative to the minimum
#'   execution time rather than absolute time.
#' @param time_unit If `NULL` the times are reported in a human readable
#'   fashion depending on each value. If one of 'ns', 'us', 'ms', 's', 'm', 'h',
#'   'd', 'w' the time units are instead expressed as nanoseconds, microseconds,
#'   milliseconds, seconds, hours, minutes, days or weeks respectively.
#' @param ... Additional arguments ignored.
#' @details
#'   If `filter_gc == TRUE` (the default) runs that contain a garbage
#'   collection will be removed before summarizing. This is most useful for fast
#'   expressions when the majority of runs do not contain a gc. Call
#'   `summary(filter_gc = FALSE)` if you would like to compute summaries _with_
#'   these times, such as expressions with lots of allocations when all or most
#'   runs contain a gc.
#' @return A [tibble][tibble::tibble] with the additional summary columns.
#'   The following summary columns are computed
#'   - `expression` - `bench_expr` The deparsed expression that was evaluated
#'     (or its name if one was provided).
#'   - `min` - `bench_time` The minimum execution time.
#'   - `median` - `bench_time` The sample median of execution time.
#'   - `itr/sec` - `double` The estimated number of executions performed per
#'   second.
#'   - `mem_alloc` - `bench_bytes` Total amount of memory allocated by R while
#'     running the expression. Memory allocated *outside* the R heap, e.g. by
#'     `malloc()` or `new` directly is *not* tracked, take care to avoid
#'     misinterpreting the results if running code that may do this.
#'   - `gc/sec` - `double` The number of garbage collections per second.
#'   - `n_itr` - `integer` Total number of iterations after filtering
#'      garbage collections (if `filter_gc == TRUE`).
#'   - `n_gc` - `double` Total number of garbage collections performed over all
#'   iterations. This is a psudo-measure of the pressure on the garbage collector, if
#'   it varies greatly between to alternatives generally the one with fewer
#'   collections will cause fewer allocation in real usage.
#'   - `total_time` - `bench_time` The total time to perform the benchmarks.
#'   - `result` - `list` A list column of the object(s) returned by the
#'     evaluated expression(s).
#'   - `memory` - `list` A list column with results from [Rprofmem()].
#'   - `time` - `list` A list column of `bench_time` vectors for each evaluated
#'     expression.
#'   - `gc` - `list` A list column with tibbles containing the level of
#'     garbage collection (0-2, columns) for each iteration (rows).
#' @examples
#' dat <- data.frame(x = runif(10000, 1, 1000), y=runif(10000, 1, 1000))
#'
#' # `bench::mark()` implicitly calls summary() automatically
#' results <- bench::mark(
#'   dat[dat$x > 500, ],
#'   dat[which(dat$x > 500), ],
#'   subset(dat, x > 500))
#'
#' # However you can also do so explicitly to filter gc differently.
#' summary(results, filter_gc = FALSE)
#'
#' # Or output relative times
#' summary(results, relative = TRUE)
#' @export
summary.bench_mark <- function(object, filter_gc = TRUE, relative = FALSE, time_unit = NULL, ...) {
  nms <- colnames(object)
  parameters <- setdiff(nms, c("expression", summary_cols, data_cols))

  num_gc <- lapply(object$gc,
    function(x) {
      res <- rowSums(x)
      if (length(res) == 0) {
        res <- rep(0, length(x))
      }
      res
    }
  )
  if (isTRUE(filter_gc)) {
    no_gc <- lapply(num_gc, `==`, 0)
    times <- Map(`[`, object$time, no_gc)
  } else {
    times <- object$time
  }

  if (filter_gc && any(lengths(times) == 0)) {
    times <- object$time
    warning(call. = FALSE,
        "Some expressions had a GC in every iteration; so filtering is disabled."
    )
  }

  object$min <- new_bench_time(vdapply(times, min))
  object$median <- new_bench_time(vdapply(times, stats::median))
  object$max <- new_bench_time(vdapply(times, max))
  object$total_time <- new_bench_time(vdapply(times, sum))

  object$n_itr <- viapply(times, length)
  object$`itr/sec` <-  as.numeric(object$n_itr / object$total_time)

  object$n_gc <- vdapply(num_gc, sum)
  object$`gc/sec` <-  as.numeric(object$n_gc / object$total_time)

  object$mem_alloc <-
    bench_bytes(
      vdapply(object$memory, function(x) if (is.null(x)) NA else sum(x$bytes, na.rm = TRUE)))

  if (isTRUE(relative)) {
    object[summary_cols] <- lapply(object[summary_cols], function(x) as.numeric(x / min(x)))
  }

  if (!is.null(time_unit)) {
    time_unit <- match.arg(time_unit, names(time_units()))
    object[time_cols] <- lapply(object[time_cols], function(x) as.numeric(x / time_units()[time_unit]))
  }

  to_keep <- intersect(c("expression", parameters, summary_cols, data_cols), names(object))
  bench_mark(object[to_keep])
}

#' @export
`[.bench_mark` <- function(x, i, j, ...) {
  bench_mark(NextMethod("["))
}

parse_allocations <- function(filename) {

  if (!file.exists(filename)) {
    empty_Rprofmem <- structure(
      list(what = character(),
        bytes = integer(),
        trace = list()),
      class = c("Rprofmem",
        "data.frame"))

    return(empty_Rprofmem)
  }

  # TODO: remove this dependency / simplify parsing
  tryCatch(
    profmem::readRprofmem(filename),
    error = function(e) {
      stop("Memory profiling failed.\n  If you are benchmarking parallel code you must set `memory = FALSE`.", call. = FALSE)
    }
  )
}

#nocov start

#' Custom printing function for bench_mark objects in knitr documents
#'
#' By default data columns ('result', 'memory', 'time', 'gc') are omitted when
#' printing in knitr. If you would like to include these columns set the knitr
#' chunk option 'bench.all_columns = TRUE'.
#' @param options A list of knitr chunk options set in the currently evaluated chunk.
#' @inheritParams knitr::knit_print
#' @details
#' You can set `bench.all_columns = TRUE` to show all columns of the bench mark
#' object.
#'
#'     ```{r bench.all_columns = TRUE}
#'     bench::mark(
#'       subset(mtcars, cyl == 3),
#'       mtcars[mtcars$cyl == 3, ])
#'     ```
knit_print.bench_mark <- function(x, ..., options) {
  if (isTRUE(options$bench.all_columns)) {
    print(x)
  } else {
    print(x[!colnames(x) %in% data_cols])
  }
}

#nocov end

parse_gc <- function(x) {
  # \x1E is Record Separator 
  x <- strsplit(paste0(x, collapse = ""), "\x1E")[[1]]
  tibble::as_tibble(.Call(parse_gc_, x))
}

utils::globalVariables(c("time", "gc"))

unnest.bench_mark <- function(data, ...) {
  if (inherits(data[["expression"]], "bench_expr")) {
    data[["expression"]] <- as.character(data[["expression"]])
  }

  # suppressWarnings to avoid 'elements may not preserve their attributes'
  # warnings from dplyr::collapse
  if (tidyr_new_interface()) {
    data <- suppressWarnings(NextMethod(.Generic, data, ...))
  } else {
    data <- suppressWarnings(NextMethod(.Generic, data, time, gc, .drop = FALSE))
  }

  # Add bench_time class back to the time column
  data$time <- as_bench_time(data$time)

  # Add a gc column, a factor with the highest gc performed for each expression.
  data$gc <-
    dplyr::case_when(
      data$level2 > 0 ~ "level2",
      data$level1 > 0 ~ "level1",
      data$level0 > 0 ~ "level0",
      TRUE ~ "none")
  data$gc <- factor(data$gc, c("none", "level0", "level1", "level2"))

  data
}

#' @export
rbind.bench_mark <- function(..., deparse.level = 1) {
  args <- list(...)
  desc <- unlist(lapply(args, function(x) as.character(x$expression)))
  res <- rbind.data.frame(...)
  attr(res$expression, "description") <- desc
  res
}

filter.bench_mark <- function(.data, ...) {
  dots <- rlang::quos(...)
  idx <- Reduce(`&`, lapply(dots, rlang::eval_tidy, data = .data))
  .data[idx, ]
}

group_by.bench_mark <- function(.data, ...) {
  bench_mark(NextMethod())
}

test_that("autoplot works", {
  skip_on_cran()
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("tidyr")
  skip_if_not_installed("ggbeeswarm")
  y <- mark(x = 1:1000)
  expect_s3_class(ggplot2::autoplot(y), "ggplot")
  expect_s3_class(plot(y), "ggplot")
})

describe("mark_", {
  it("If min_time is Inf, runs for max_iterations", {
    res <- .Call(
      mark_,
      quote(1),
      new.env(),
      Inf,
      as.integer(0),
      as.integer(10),
      FALSE
    )
    expect_length(res, 10)

    res <- .Call(
      mark_,
      quote(1),
      new.env(),
      Inf,
      as.integer(0),
      as.integer(20),
      FALSE
    )
    expect_length(res, 20)
  })

  it("If min_time is 0, runs for min_iterations", {
    res <- .Call(
      mark_,
      quote(1),
      new.env(),
      0,
      as.integer(1),
      as.integer(10),
      FALSE
    )
    expect_length(res, 1)

    res <- .Call(
      mark_,
      quote(1),
      new.env(),
      0,
      as.integer(5),
      as.integer(10),
      FALSE
    )
    expect_length(res, 5)
  })

  it("If min_time is 0, runs for min_iterations", {
    res <- .Call(
      mark_,
      quote({
        i <- 1
        while (i < 10000) i <- i + 1
      }),
      new.env(),
      .1,
      as.integer(1),
      as.integer(1000),
      FALSE
    )

    expect_gte(length(res), 1)
    expect_lte(length(res), 1000)
  })

  it("Evaluates code in the environment", {
    e <- new.env(parent = baseenv())
    res <- .Call(
      mark_,
      quote({
        a <- 42
      }),
      e,
      Inf,
      as.integer(1),
      as.integer(1),
      FALSE
    )
    expect_equal(e[["a"]], 42)
  })
})

describe("mark", {
  it("Uses all.equal to check results by default", {
    res <- mark(1 + 1, 1L + 1L, check = TRUE, iterations = 1)

    expect_type(res$result, "list")
    expect_true(all.equal(res$result[[1]], res$result[[2]]))
  })
  it("Can use other functions to check results like identical to check results", {
    # numerics and integers not identical
    expect_error(
      regexp = "Each result must equal the first result",
      mark(1 + 1, 1L + 1L, check = identical, iterations = 1)
    )

    # Function that always returns false
    expect_error(
      regexp = "Each result must equal the first result",
      mark(1 + 1, 1 + 1, check = function(x, y) FALSE, iterations = 1)
    )

    # Function that always returns true
    res <- mark(1 + 1, 1 + 2, check = function(x, y) TRUE, iterations = 1)

    expect_type(res$result, "list")
    expect_equal(res$result[[1]], 2)
    expect_equal(res$result[[2]], 3)
  })

  it("works with capabilities('profmem')", {
    skip_if_not(capabilities("profmem"))

    res <- mark(1, 2, check = FALSE, iterations = 1)

    expect_length(res$memory, 2)

    expect_s3_class(res$memory[[1]], "Rprofmem")
    expect_equal(ncol(res$memory[[1]]), 3)
    expect_gte(nrow(res$memory[[1]]), 0)
  })

  it("works without capabilities('profmem')", {
    res <- mark(1, 2, check = FALSE, iterations = 1, memory = FALSE)

    expect_equal(res$memory, vector("list", 2))
    expect_equal(res$mem_alloc, as_bench_bytes(c(NA, NA)))
  })
  it("Can handle `NULL` results", {
    res <- mark(if (FALSE) 1, max_iterations = 10)
    expect_equal(res$result, list(NULL))
  })
  it("Can errors with the deparsed expressions", {
    expect_snapshot(error = TRUE, {
      mark(1, 1, 3, max_iterations = 10)
    })
  })

  it("Works when calls are different lengths", {
    expect_snapshot(error = TRUE, {
      # Here the first call deparses to length 2, the second to length 4
      mark(if (TRUE) 2, if (TRUE) 1 else 3)
    })
  })
  it("works with memory = FALSE", {
    res <- mark(1, memory = FALSE)
    expect_s3_class(res, "bench_mark")
    expect_equal(res$memory, vector("list", 1))
    expect_equal(res$mem_alloc, as_bench_bytes(NA))
  })
  it("works with check = FALSE", {
    res <- mark(1, check = FALSE)
    expect_s3_class(res, "bench_mark")
    expect_equal(res$result, list(NULL))
  })
  it("works with memory = FALSE and check = FALSE", {
    res <- mark(1, memory = FALSE, check = FALSE)
    expect_s3_class(res, "bench_mark")
    expect_equal(res$memory, list(NULL))
    expect_equal(res$mem_alloc, as_bench_bytes(NA))
  })
  it("fails for memory profiling failures", {
    skip_on_os("windows")
    skip_on_cran()

    keep_busy <- function(n = 1e3) {
      r <- rnorm(n)
      p <- pnorm(r)
      q <- qnorm(p)
      o <- order(q)
    }
    expect_error(
      res <- mark(parallel::mclapply(seq_len(1e3), keep_busy, mc.cores = 2)),
      "Memory profiling failed"
    )
  })
  it("ignores trailing arguments", {
    bench::mark(
      1 + 3,
      2 + 2,
    )
  })
  it("truncates long expressions when printing (#94)", {
    local_reproducible_output(width = 30)

    name <- paste0(rep("a", 100), collapse = "")
    exprs <- list(as.name(name))

    assign(name, 1, envir = environment())

    out <- mark(exprs = exprs)

    # Only snapshot static columns
    out <- out[c("expression", "result")]

    expect_snapshot(out)
  })
})

describe("summary.bench_mark", {
  res <- bench_mark(
    tibble::tibble(
      expression = "1 + 1:1e+06",
      result = list(1:10),
      memory = list(NULL),
      time = list(
        c(
          0.088492998,
          0.109396977,
          0.141906863,
          0.005378346,
          0.007563524,
          0.002439451,
          0.079715252,
          0.003022223,
          0.005948069,
          0.002276121
        )
      ),
      gc = list(
        tibble::tibble(
          level0 = c(1, 0, 0, 0, 1, 0, 0, 0, 1, 0),
          level1 = c(0, 1, 0, 0, 0, 0, 0, 0, 0, 0),
          level2 = c(0, 0, 1, 0, 0, 0, 1, 0, 0, 0)
        )
      )
    )
  )
  it("computes relative summaries if called with relative = TRUE", {
    # remove memory column, as there likely are no allocations or gc in these
    # benchmarks
    res1 <- summary(res)
    for (col in setdiff(summary_cols, "mem_alloc")) {
      # Absolute values should always be positive
      expect_true(all(res1[[!!col]] >= 0))
    }

    # Relative values should always be greater than or equal to 1
    res2 <- summary(res, relative = TRUE)
    for (col in setdiff(summary_cols, c("mem_alloc", "n_gc"))) {
      expect_true(all(res2[[!!col]] >= 1))
    }
  })
  it("does not filter gc is `filter_gc` is FALSE", {
    res1 <- summary(res, filter_gc = TRUE)
    res2 <- summary(res, filter_gc = FALSE)

    expect_equal(res1$n_gc, 6)
    expect_equal(res1$n_gc, res2$n_gc)
  })

  it("does not issue warnings if there are no garbage collections", {
    # This is artificial, but it avoids differences in gc on different
    # platforms / memory loads, so we can ensure the first has no gcs, and the
    # second has all gcs
    x <- bench_mark(
      tibble::tibble(
        expression = c(1, 2),
        result = list(1, 2),
        time = list(
          as_bench_time(c(0.166, 0.161, 0.162)),
          as_bench_time(c(0.276, 0.4))
        ),
        memory = list(NULL, NULL),
        gc = list(
          tibble::tibble(
            level0 = integer(0),
            level1 = integer(0),
            level2 = integer(0)
          ),
          tibble::tibble(
            level0 = c(1L, 1L),
            level1 = c(0L, 0L),
            level2 = c(0L, 0L)
          )
        )
      )
    )

    expect_warning(
      regexp = "Some expressions had a GC in every iteration",
      res <- summary(x, filter_gc = TRUE)
    )

    expect_equal(res$min, as_bench_time(c(.161, .276)))
    expect_equal(res$median, as_bench_time(c(.162, .338)))
    expect_equal(res$`itr/sec`, c(6.134969, 2.958580), tolerance = 1e-5)
    expect_equal(res$mem_alloc, as_bench_bytes(c(NA, NA)))
    expect_equal(res$`gc/sec`, c(0, 2.958580), tolerance = 1e-5)
    expect_equal(res$n_gc, c(0, 2))
    expect_equal(res$n_itr, c(3, 2))
    expect_equal(res$total_time, as_bench_time(c(.489, .676)))

    expect_no_warning(res2 <- summary(x, filter_gc = FALSE))

    expect_identical(res, res2)
  })
})

describe("unnest.bench_mark", {
  it("does not contain result or memory columns", {
    skip_if_not_installed("tidyr", "1.0.0")
    bnch <- mark(1 + 1, 2 + 0)
    res <- tidyr::unnest(bnch, c(time, gc))

    gc_cols <- colnames(bnch$gc[[1]])

    expect_equal(
      colnames(res),
      c(head(colnames(bnch), n = -1), c(gc_cols, "gc"))
    )

    expect_equal(nrow(res), length(bnch$time[[1]]) + length(bnch$time[[2]]))
  })
})

## usethis namespace: start
#' @importFrom methods setOldClass
#' @importFrom pillar pillar_shaft
#' @importFrom pillar type_sum
#' @importFrom rlang .data
## usethis namespace: end
NULL

#' @keywords internal
#' @inherit summary.bench_mark examples
"_PACKAGE"

# mark: Can errors with the deparsed expressions

    Code
      mark(1, 1, 3, max_iterations = 10)
    Condition
      Error:
      ! Each result must equal the first result:
      `1` does not equal `3`

# mark: Works when calls are different lengths

    Code
      mark(if (TRUE) 2, if (TRUE) 1 else 3)
    Condition
      Error:
      ! Each result must equal the first result:
      `if (TRUE) 2` does not equal `if (TRUE) 1 else 3`

# mark: truncates long expressions when printing (#94)

    Code
      out
    Output
      # A tibble: 1 x 2
        expression            result
        <bch:expr>            <list>
      1 aaaaaaaaaaaaaaaaaaaa~ <dbl> 

# press: Outputs status message before evaluating each parameter

    Code
      res <- press(x = 1, mark(rep(1, x), max_iterations = 10))
    Message
      Running with:
            x
      1     1

---

    Code
      messages
    Output
      [1] "Running with:\n      x\n" "1     1\n"               
      [3] "2     2\n"                "3     3\n"               

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/autoplot.R
\name{autoplot.bench_mark}
\alias{autoplot.bench_mark}
\alias{plot.bench_mark}
\title{Autoplot method for bench_mark objects}
\usage{
autoplot.bench_mark(
  object,
  type = c("beeswarm", "jitter", "ridge", "boxplot", "violin"),
  ...
)

\method{plot}{bench_mark}(x, ..., type = c("beeswarm", "jitter", "ridge", "boxplot", "violin"), y)
}
\arguments{
\item{object}{A \code{bench_mark} object.}

\item{type}{The type of plot. Plotting geoms used for each type are
\itemize{
\item beeswarm - \code{\link[ggbeeswarm:geom_quasirandom]{ggbeeswarm::geom_quasirandom()}}
\item jitter - \code{\link[ggplot2:geom_jitter]{ggplot2::geom_jitter()}}
\item ridge - \code{\link[ggridges:geom_density_ridges]{ggridges::geom_density_ridges()}}
\item boxplot - \code{\link[ggplot2:geom_boxplot]{ggplot2::geom_boxplot()}}
\item violin - \code{\link[ggplot2:geom_violin]{ggplot2::geom_violin()}}
}}

\item{...}{Additional arguments passed to the plotting geom.}

\item{x}{A \code{bench_mark} object.}

\item{y}{Ignored, required for compatibility with the \code{plot()} generic.}
}
\description{
Autoplot method for bench_mark objects
}
\details{
This function requires some optional dependencies. \link[ggplot2:ggplot2-package]{ggplot2},
\link[tidyr:tidyr-package]{tidyr}, and depending on the plot type
\link[ggbeeswarm:ggbeeswarm]{ggbeeswarm}, \link[ggridges:ggridges]{ggridges}.

For \code{type} of \code{beeswarm} and \code{jitter} the points are colored by the highest
level garbage collection performed during each iteration.

For plots with 2 parameters \code{ggplot2::facet_grid()} is used to construct a
2d facet. For other numbers of parameters \code{ggplot2::facet_wrap()} is used
instead.
}
\examples{
dat <- data.frame(x = runif(10000, 1, 1000), y=runif(10000, 1, 1000))

res <- bench::mark(
  dat[dat$x > 500, ],
  dat[which(dat$x > 500), ],
  subset(dat, x > 500))

if (require(ggplot2) && require(tidyr)) {

  # Beeswarm plot
  autoplot(res)

  # ridge (joyplot)
  autoplot(res, "ridge")

  # If you want to have the plots ordered by execution time you can do so by
  # ordering factor levels in the expressions.
  if (require(dplyr) && require(forcats)) {

    res \%>\%
      mutate(expression = forcats::fct_reorder(as.character(expression), min, .desc = TRUE)) \%>\%
      as_bench_mark() \%>\%
      autoplot("violin")
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mark.R
\name{knit_print.bench_mark}
\alias{knit_print.bench_mark}
\title{Custom printing function for \code{bench_mark} objects in knitr documents}
\usage{
knit_print.bench_mark(x, ..., options)
}
\arguments{
\item{x}{An R object to be printed}

\item{...}{Additional arguments passed to the S3 method. Currently ignored,
except two optional arguments \code{options} and \code{inline}; see
the references below.}

\item{options}{A list of knitr chunk options set in the currently evaluated
chunk.}
}
\description{
By default, data columns (\code{result}, \code{memory}, \code{time}, \code{gc}) are omitted when
printing in knitr. If you would like to include these columns, set the knitr
chunk option \code{bench.all_columns = TRUE}.
}
\details{
You can set \code{bench.all_columns = TRUE} to show all columns of the bench mark
object.

\if{html}{\out{<div class="sourceCode">}}\preformatted{```\{r, bench.all_columns = TRUE\}
bench::mark(
  subset(mtcars, cyl == 3),
  mtcars[mtcars$cyl == 3, ]
)
```
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bench-package.R
\docType{package}
\name{bench-package}
\alias{bench}
\alias{bench-package}
\title{bench: High Precision Timing of R Expressions}
\description{
Tools to accurately benchmark and analyze execution times for R expressions.
}
\examples{
dat <- data.frame(x = runif(10000, 1, 1000), y=runif(10000, 1, 1000))

# `bench::mark()` implicitly calls summary() automatically
results <- bench::mark(
  dat[dat$x > 500, ],
  dat[which(dat$x > 500), ],
  subset(dat, x > 500))

# However you can also do so explicitly to filter gc differently.
summary(results, filter_gc = FALSE)

# Or output relative times
summary(results, relative = TRUE)
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/r-lib/bench}
  \item Report bugs at \url{https://github.com/r-lib/bench/issues}
}

}
\author{
\strong{Maintainer}: Jim Hester \email{james.f.hester@gmail.com}

Other contributors:
\itemize{
  \item Drew Schmidt (read_proc_file implementation) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hires_time.R
\name{hires_time}
\alias{hires_time}
\title{Return the current high-resolution real time.}
\usage{
hires_time()
}
\description{
Time is expressed as seconds since some arbitrary time in the past; it
is not correlated in any way to the time of day, and thus is not subject to
resetting or drifting. The hi-res
timer is ideally suited to performance measurement tasks, where cheap,
accurate interval timing is required.
}
\examples{
hires_time()

# R rounds doubles to 7 digits by default, see greater precision by setting
# the digits argument when printing
print(hires_time(), digits = 20)

# Generally used by recording two times and then subtracting them
start <- hires_time()
end <- hires_time()
elapsed <- end - start
elapsed
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bench_time.R
\name{bench_time}
\alias{bench_time}
\alias{system_time}
\title{Measure Process CPU and real time that an expression used.}
\usage{
bench_time(expr)
}
\arguments{
\item{expr}{A expression to be timed.}
}
\value{
A \code{bench_time} object with two values.
\itemize{
\item \code{process} - The process CPU usage of the expression evaluation.
\item \code{real} - The wallclock time of the expression evaluation.
}
}
\description{
Measure Process CPU and real time that an expression used.
}
\details{
On some systems (such as macOS) the process clock has lower
precision than the realtime clock, as a result there may be cases where the
process time is larger than the real time for fast expressions.
}
\examples{
# This will use ~.5 seconds of real time, but very little process time.
bench_time(Sys.sleep(.5))
}
\seealso{
\code{\link[=bench_memory]{bench_memory()}} To measure memory allocations for a given expression.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bench_process_memory.R
\name{bench_process_memory}
\alias{bench_process_memory}
\title{Retrieve the current and maximum memory from the R process}
\usage{
bench_process_memory()
}
\description{
The memory reported here will likely differ from that reported by \code{gc()}, as
this includes all memory from the R process, including any child processes
and memory allocated outside R's garbage collector heap.
}
\details{
The OS APIs used are as follows
\subsection{Windows}{
\itemize{
\item \href{https://docs.microsoft.com/en-us/windows/win32/api/psapi/ns-psapi-process_memory_counters}{PROCESS_MEMORY_COUNTERS.WorkingSetSize}
\item \href{https://docs.microsoft.com/en-us/windows/win32/api/psapi/ns-psapi-process_memory_counters}{PROCESS_MEMORY_COUNTERS.PeakWorkingSetSize}
}
}

\subsection{macOS}{
\itemize{
\item \href{https://developer.apple.com/documentation/kernel/1537934-task_info?language=occ}{task_info(TASK_BASIC_INFO)}
\item \href{https://developer.apple.com/library/archive/documentation/System/Conceptual/ManPages_iPhoneOS/man2/getrusage.2.html}{rusage.ru_maxrss}
}
}

\subsection{linux}{
\itemize{
\item \href{http://man7.org/linux/man-pages/man5/proc.5.html}{/proc/pid/status VmSize}
\item \href{http://man7.org/linux/man-pages/man5/proc.5.html}{/proc/pid/status VmPeak}
and on Windows
\href{https://docs.microsoft.com/en-us/windows/win32/api/psapi/ns-psapi-process_memory_counters}{PROCESS_MEMORY_COUNTERS.PeakWorkingSetSize}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load.R
\name{bench_load_average}
\alias{bench_load_average}
\title{Get system load averages}
\usage{
bench_load_average()
}
\description{
Uses OS system APIs to return the load average for the past 1, 5 and 15 minutes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/press.R
\name{press}
\alias{press}
\title{Run setup code and benchmarks across a grid of parameters}
\usage{
press(..., .grid = NULL)
}
\arguments{
\item{...}{If named, parameters to define, if unnamed the expression to run.
Only one unnamed expression is permitted.}

\item{.grid}{A pre-built grid of values to use, typically a \link{data.frame} or
\link{tibble}. This is useful if you only want to benchmark a subset of all
possible combinations.}
}
\description{
\code{press()} is used to run \code{\link[=mark]{mark()}} across a grid of parameters and
then \emph{press} the results together.

The parameters you want to set are given as named arguments and a grid of
all possible combinations is automatically created.

The code to setup and benchmark is given by one unnamed expression (often
delimited by \verb{\\\{}).

If replicates are desired a dummy variable can be used, e.g. \code{rep = 1:5} for
replicates.
}
\examples{
# Helper function to create a simple data.frame of the specified dimensions
create_df <- function(rows, cols) {
  as.data.frame(setNames(
    replicate(cols, runif(rows, 1, 1000), simplify = FALSE),
    rep_len(c("x", letters), cols)))
}

# Run 4 data sizes across 3 samples with 2 replicates (24 total benchmarks)
press(
  rows = c(1000, 10000),
  cols = c(10, 100),
  rep = 1:2,
  {
    dat <- create_df(rows, cols)
    bench::mark(
      min_time = .05,
      bracket = dat[dat$x > 500, ],
      which = dat[which(dat$x > 500), ],
      subset = subset(dat, x > 500)
    )
  }
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mark.R
\name{as_bench_mark}
\alias{as_bench_mark}
\title{Coerce to a bench mark object Bench mark objects}
\usage{
as_bench_mark(x)
}
\arguments{
\item{x}{Object to be coerced}
}
\description{
This is typically needed only if you are performing additional manipulations
after calling \code{\link[=mark]{mark()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mark.R
\name{summary.bench_mark}
\alias{summary.bench_mark}
\title{Summarize \link[bench:mark]{bench::mark} results.}
\usage{
\method{summary}{bench_mark}(object, filter_gc = TRUE, relative = FALSE, time_unit = NULL, ...)
}
\arguments{
\item{object}{\link{bench_mark} object to summarize.}

\item{filter_gc}{If \code{TRUE} remove iterations that contained at least one
garbage collection before summarizing. If \code{TRUE} but an expression had
a garbage collection in every iteration, filtering is disabled, with a warning.}

\item{relative}{If \code{TRUE} all summaries are computed relative to the minimum
execution time rather than absolute time.}

\item{time_unit}{If \code{NULL} the times are reported in a human readable
fashion depending on each value. If one of 'ns', 'us', 'ms', 's', 'm', 'h',
'd', 'w' the time units are instead expressed as nanoseconds, microseconds,
milliseconds, seconds, hours, minutes, days or weeks respectively.}

\item{...}{Additional arguments ignored.}
}
\value{
A \link[tibble:tibble]{tibble} with the additional summary columns.
The following summary columns are computed
\itemize{
\item \code{expression} - \code{bench_expr} The deparsed expression that was evaluated
(or its name if one was provided).
\item \code{min} - \code{bench_time} The minimum execution time.
\item \code{median} - \code{bench_time} The sample median of execution time.
\item \code{itr/sec} - \code{double} The estimated number of executions performed per
second.
\item \code{mem_alloc} - \code{bench_bytes} Total amount of memory allocated by R while
running the expression. Memory allocated \emph{outside} the R heap, e.g. by
\code{malloc()} or \code{new} directly is \emph{not} tracked, take care to avoid
misinterpreting the results if running code that may do this.
\item \code{gc/sec} - \code{double} The number of garbage collections per second.
\item \code{n_itr} - \code{integer} Total number of iterations after filtering
garbage collections (if \code{filter_gc == TRUE}).
\item \code{n_gc} - \code{double} Total number of garbage collections performed over all
iterations. This is a psudo-measure of the pressure on the garbage collector, if
it varies greatly between to alternatives generally the one with fewer
collections will cause fewer allocation in real usage.
\item \code{total_time} - \code{bench_time} The total time to perform the benchmarks.
\item \code{result} - \code{list} A list column of the object(s) returned by the
evaluated expression(s).
\item \code{memory} - \code{list} A list column with results from \code{\link[=Rprofmem]{Rprofmem()}}.
\item \code{time} - \code{list} A list column of \code{bench_time} vectors for each evaluated
expression.
\item \code{gc} - \code{list} A list column with tibbles containing the level of
garbage collection (0-2, columns) for each iteration (rows).
}
}
\description{
Summarize \link[bench:mark]{bench::mark} results.
}
\details{
If \code{filter_gc == TRUE} (the default) runs that contain a garbage
collection will be removed before summarizing. This is most useful for fast
expressions when the majority of runs do not contain a gc. Call
\code{summary(filter_gc = FALSE)} if you would like to compute summaries \emph{with}
these times, such as expressions with lots of allocations when all or most
runs contain a gc.
}
\examples{
dat <- data.frame(x = runif(10000, 1, 1000), y=runif(10000, 1, 1000))

# `bench::mark()` implicitly calls summary() automatically
results <- bench::mark(
  dat[dat$x > 500, ],
  dat[which(dat$x > 500), ],
  subset(dat, x > 500))

# However you can also do so explicitly to filter gc differently.
summary(results, filter_gc = FALSE)

# Or output relative times
summary(results, relative = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workout.R
\name{workout}
\alias{workout}
\alias{workout_expressions}
\title{Workout a group of expressions individually}
\usage{
workout(expr, description = NULL)

workout_expressions(exprs, env = parent.frame(), description = NULL)
}
\arguments{
\item{expr}{one or more expressions to workout, use \code{{}} to pass multiple
expressions.}

\item{description}{A name to label each expression, if not supplied the
deparsed expression will be used.}

\item{exprs}{A list of calls to measure.}

\item{env}{The environment in which the expressions should be evaluated.}
}
\description{
Given an block of expressions in \code{{}} \code{\link[=workout]{workout()}} individually times each
expression in the group. \code{\link[=workout_expressions]{workout_expressions()}} is a lower level function most
useful when reading lists of calls from a file.
}
\examples{
workout({
  x <- 1:1000
  evens <- x \%\% 2 == 0
  y <- x[evens]
  length(y)
  length(which(evens))
  sum(evens)
})

# The equivalent to the above, reading the code from a file
workout_expressions(as.list(parse(system.file("examples/exprs.R", package = "bench"))))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bytes.R
\name{bench_bytes}
\alias{bench_bytes}
\alias{as_bench_bytes}
\title{Human readable memory sizes}
\usage{
as_bench_bytes(x)

bench_bytes(x)
}
\arguments{
\item{x}{A numeric or character vector. Character representations can use
shorthand sizes (see examples).}
}
\description{
Construct, manipulate and display vectors of byte sizes. These are numeric
vectors, so you can compare them numerically, but they can also be compared
to human readable values such as '10MB'.
}
\details{
These memory sizes are always assumed to be base 1024, rather than 1000.
}
\examples{
bench_bytes("1")
bench_bytes("1K")
bench_bytes("1Kb")
bench_bytes("1KiB")
bench_bytes("1MB")

bench_bytes("1KB") < "1MB"

sum(bench_bytes(c("1MB", "5MB", "500KB")))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expression.R
\name{scale_bench_expr}
\alias{scale_bench_expr}
\alias{scale_x_bench_expr}
\alias{scale_y_bench_expr}
\alias{scale_colour_bench_expr}
\alias{scale_color_bench_expr}
\title{Position and color scales for bench_expr data}
\usage{
scale_x_bench_expr(...)

scale_y_bench_expr(...)

scale_colour_bench_expr(
  palette = scales::hue_pal(...),
  ...,
  aesthetics = "colour"
)

scale_color_bench_expr(
  palette = scales::hue_pal(...),
  ...,
  aesthetics = "colour"
)
}
\description{
Default scales for the \code{bench_expr} class, these are added to plots using
\code{bench_expr} objects automatically.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time.R
\name{bench_time_trans}
\alias{bench_time_trans}
\title{Benchmark time transformation}
\usage{
bench_time_trans(base = 10)
}
\arguments{
\item{base}{base of logarithm}
}
\description{
This both log transforms the times and formats the labels as a \code{bench_time}
object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bytes.R, R/time.R
\name{scale_bench_time}
\alias{scale_bench_time}
\alias{scale_x_bench_bytes}
\alias{scale_y_bench_bytes}
\alias{scale_x_bench_time}
\alias{scale_y_bench_time}
\title{Position scales for bench_time data}
\usage{
scale_x_bench_bytes(base = 10, ...)

scale_y_bench_bytes(base = 10, ...)

scale_x_bench_time(base = 10, ...)

scale_y_bench_time(base = 10, ...)
}
\arguments{
\item{base}{The base of the logarithm, if \code{NULL} instead use a
non-logarithmic scale.}
}
\description{
Default scales for the \code{bench_time} class, these are added to plots using
\code{bench_time} objects automatically.

Default scales for the \code{bench_time} class, these are added to plots using
\code{bench_time} objects automatically.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bench_time.R
\name{bench_memory}
\alias{bench_memory}
\title{Measure memory that an expression used.}
\usage{
bench_memory(expr)
}
\arguments{
\item{expr}{A expression to be measured.}
}
\value{
A tibble with two columns
\itemize{
\item The total amount of memory allocated
\item The raw memory allocations as parsed by \code{\link[profmem:readRprofmem]{profmem::readRprofmem()}}
}
}
\description{
Measure memory that an expression used.
}
\examples{
if (capabilities("profmem")) {
  bench_memory(1 + 1:10000)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mark.R
\name{mark}
\alias{mark}
\alias{bench_mark}
\title{Benchmark a series of functions}
\usage{
mark(
  ...,
  min_time = 0.5,
  iterations = NULL,
  min_iterations = 1,
  max_iterations = 10000,
  check = TRUE,
  memory = capabilities("profmem"),
  filter_gc = TRUE,
  relative = FALSE,
  time_unit = NULL,
  exprs = NULL,
  env = parent.frame()
)
}
\arguments{
\item{...}{Expressions to benchmark, if named the \code{expression} column will
be the name, otherwise it will be the deparsed expression.}

\item{min_time}{The minimum number of seconds to run each expression, set to
\code{Inf} to always run \code{max_iterations} times instead.}

\item{iterations}{If not \code{NULL}, the default, run each expression for
exactly this number of iterations. This overrides both \code{min_iterations}
and \code{max_iterations}.}

\item{min_iterations}{Each expression will be evaluated a minimum of \code{min_iterations} times.}

\item{max_iterations}{Each expression will be evaluated a maximum of \code{max_iterations} times.}

\item{check}{Check if results are consistent. If \code{TRUE}, checking is done
with \code{\link[=all.equal]{all.equal()}}, if \code{FALSE} checking is disabled and results are not
stored. If \code{check} is a function that function will be called with each
pair of results to determine consistency.}

\item{memory}{If \code{TRUE} (the default when R is compiled with memory
profiling), track memory allocations using. If \code{FALSE} disable memory
tracking.}

\item{filter_gc}{If \code{TRUE} remove iterations that contained at least one
garbage collection before summarizing. If \code{TRUE} but an expression had
a garbage collection in every iteration, filtering is disabled, with a warning.}

\item{relative}{If \code{TRUE} all summaries are computed relative to the minimum
execution time rather than absolute time.}

\item{time_unit}{If \code{NULL} the times are reported in a human readable
fashion depending on each value. If one of 'ns', 'us', 'ms', 's', 'm', 'h',
'd', 'w' the time units are instead expressed as nanoseconds, microseconds,
milliseconds, seconds, hours, minutes, days or weeks respectively.}

\item{exprs}{A list of quoted expressions. If supplied overrides expressions
defined in \code{...}.}

\item{env}{The environment which to evaluate the expressions}
}
\value{
A \link[tibble:tibble]{tibble} with the additional summary columns.
The following summary columns are computed
\itemize{
\item \code{expression} - \code{bench_expr} The deparsed expression that was evaluated
(or its name if one was provided).
\item \code{min} - \code{bench_time} The minimum execution time.
\item \code{median} - \code{bench_time} The sample median of execution time.
\item \code{itr/sec} - \code{double} The estimated number of executions performed per
second.
\item \code{mem_alloc} - \code{bench_bytes} Total amount of memory allocated by R while
running the expression. Memory allocated \emph{outside} the R heap, e.g. by
\code{malloc()} or \code{new} directly is \emph{not} tracked, take care to avoid
misinterpreting the results if running code that may do this.
\item \code{gc/sec} - \code{double} The number of garbage collections per second.
\item \code{n_itr} - \code{integer} Total number of iterations after filtering
garbage collections (if \code{filter_gc == TRUE}).
\item \code{n_gc} - \code{double} Total number of garbage collections performed over all
iterations. This is a psudo-measure of the pressure on the garbage collector, if
it varies greatly between to alternatives generally the one with fewer
collections will cause fewer allocation in real usage.
\item \code{total_time} - \code{bench_time} The total time to perform the benchmarks.
\item \code{result} - \code{list} A list column of the object(s) returned by the
evaluated expression(s).
\item \code{memory} - \code{list} A list column with results from \code{\link[=Rprofmem]{Rprofmem()}}.
\item \code{time} - \code{list} A list column of \code{bench_time} vectors for each evaluated
expression.
\item \code{gc} - \code{list} A list column with tibbles containing the level of
garbage collection (0-2, columns) for each iteration (rows).
}
}
\description{
Benchmark a list of quoted expressions. Each expression will always run at
least twice, once to measure the memory allocation and store results and one
or more times to measure timing.
}
\examples{
dat <- data.frame(x = runif(100, 1, 1000), y=runif(10, 1, 1000))
mark(
  min_time = .1,

  dat[dat$x > 500, ],
  dat[which(dat$x > 500), ],
  subset(dat, x > 500))
}
\seealso{
\code{\link[=press]{press()}} to run benchmarks across a grid of parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time.R
\name{as_bench_time}
\alias{as_bench_time}
\title{Human readable times}
\usage{
as_bench_time(x)
}
\arguments{
\item{x}{A numeric or character vector. Character representations can use
shorthand sizes (see examples).}
}
\description{
Construct, manipulate and display vectors of elapsed times in seconds. These
are numeric vectors, so you can compare them numerically, but they can also
be compared to human readable values such as '10ms'.
}
\examples{
as_bench_time("1ns")
as_bench_time("1")
as_bench_time("1us")
as_bench_time("1ms")
as_bench_time("1s")

as_bench_time("100ns") < "1ms"

sum(as_bench_time(c("1MB", "5MB", "500KB")))
}
